
package org.mdanderson.biostat.geneclust;

import java.io.BufferedWriter;
import java.io.IOException;

final class DisplayProcess extends SProcess
{
    // =========================================================================
    // CONSTRUCTORS
    // =========================================================================

    /**
     * Class constructor specifying (way too many) simulation parameters...
     *
     * @param dialog                  dialog allows monitoring of sim progress
     * @param isDisplayGenesGraph     if true, display genes graph
     * @param isDisplaySamplesGraph   if true, display samples graph
     * @param isDisplayDataMatrix   if true, display samples graph
     * @param isDisplayShaving   if true, display samples graph
     * @throws NullPointerException if dialog is null
     */
    public DisplayProcess(SimulationDialog dialog,
	    	          String outputName,
                      boolean isDisplayGenesGraph,
                      boolean isDisplaySamplesGraph,
                      boolean isDisplayDataMatrix,
                      boolean isDisplayShaving)
    {
	super(dialog);

        this.outputName = outputName;
        this.displayGenes = isDisplayGenesGraph;
        this.displaySamples = isDisplaySamplesGraph;
        this.displayDataMatrix = isDisplayDataMatrix;
        this.displayShaving = isDisplayShaving;
    }

    // =========================================================================
    // PROTECTED INTERFACE
    // =========================================================================

    /**
     * Writes the S interpreter initialization code. It sets S interpreter options and global
     * variables, dynamically loads the C library, and initializes the S interpreter
     * graphics environment.
     *
     * @param bw  output stream writer
     * @throws IOException if an error occurs while writing
     */
    protected void writeSinterpreterInit(SBufferedWriter bw)
      throws IOException
    {
        bw.Command ("options(echo=TRUE);");
	bw.SetStringVar ("GCinstall",     installDirName);
	bw.SetStringVar ("GCinterpreter", interpreterName);
	bw.SetStringVar ("GCoutput",      outputDirName);

	bw.Source ("GCObjectLoader.s");
	bw.Source ("DisplayCluster.s");
	bw.Source ("GeneClust.s");
        bw.Command("options(error=ErrorHandler.func);");

	bw.Source ("Graphics.s");
        bw.Command("InitGraphics.func();");

        bw.write("cat(\"PK_MONITOR_APP_PROCESSING_INPUT\\n\")");
        bw.newLine();

        bw.flush();
    }

    /**
     * Writes mode-indifferent S interpreter input data to the output stream.
     *
     * @param bw  output stream writer
     * @throws IOException if an error occurs while writing
     */
    protected void writeSinterpreterGenericInput(SBufferedWriter bw)
      throws IOException
    {
        bw.SetLogicalVar("graphGenes",  displayGenes);
        bw.SetLogicalVar("graphSamples", displaySamples);
        bw.SetLogicalVar("graphDataMatrix", displayDataMatrix);
        bw.SetLogicalVar("graphShaving", displayShaving);
	bw.SetStringVar ("displayName",   outputName);
        bw.Command("GCDisplayData.func();");
    }

    /**
     * Writes all S interpreter simulation input to the output stream.
     *
     * @param bw  output stream writer
     */
    public void writeSinterpreterInput(SBufferedWriter bw)
    {
        try
        {
            synchronized(bw)
            {
		bw.TimeStamp ("GeneClust rawdata input");
                writeSinterpreterInit(bw);
                writeSinterpreterGenericInput(bw);
                bw.flush();
            }
        }
        catch (IOException ioe)
        {
            System.err.println(ioe);
        }
    }

    // =========================================================================
    // CLASS ATTRIBUTES
    // =========================================================================

    private static String installDirName;
    private static String interpreterName;
    private static String outputDirName;

    static
    {
        final String keyGCInstallDir = "gcinstall.home";
        final String keyGCOutputDir = "geneclust.output";
        final String keyGCInterpreterName = "sinterpreter.name";

        installDirName = System.getProperty(keyGCInstallDir);
        interpreterName = System.getProperty(keyGCInterpreterName);
        outputDirName = System.getProperty(keyGCOutputDir);
    }

    // =========================================================================
    // OBJECT ATTRIBUTES
    // =========================================================================

    protected String outputName;
    protected boolean displayGenes;
    protected boolean displaySamples;
    protected boolean displayDataMatrix;
    protected boolean displayShaving;
}
