/*=================== FILE ABSTRACT ==========================================
 *
 * Synopsis::         Source for GeneClust user interface (StreamMonitor)
 *
 * File Name::        %M%
 * Version::          %I%
 * Type::             source
 * Date Completed::   %G%
 *
 * Environment::
 *     Language::     Java
 *     Computer::     Compaq/Alpha
 *     OS::           DigitalUNIX
 *     OSVersion::    4.0
 *
 * Module Information::
 *     Owner::        MDACC
 *     Department::   Biostatistics
 *     Researcher::   Kim-Anh Do
 *     Application::  GeneClust
 *
 *============================================================================*/

/*
 * @(#)GCStreamMonitor.java    1.0 2001/08/16
 *
 * Copyright (c) 2001 M.D. Anderson Cancer Center
 * 1515 Holcombe, Houston, TX, 77030, U.S.A.
 * All rights reserved.
 */

package org.mdanderson.biostat.geneclust;

import java.io.InputStream;
import java.io.OutputStream;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;
import java.util.StringTokenizer;
import javax.swing.SwingUtilities;
import org.mdanderson.biostat.util.DefaultStreamMonitor;
import org.mdanderson.biostat.util.StreamMonitor;


/**
 * GCStreamMonitor - Implements application-specific stream monitoring
 *
 * @version 1.0 2001/08/16
 * @author <a href="mailto:roebuck@mdanderson.org">P. Roebuck</a>
 */
public class GCStreamMonitor extends DefaultStreamMonitor
    implements GCStreamMonitorKeys
{
    // =========================================================================
    // CONSTRUCTOR
    // =========================================================================

    /**
     * Class constructor specifying input stream, output stream, and dialog
     * where results should be displayed
     *
     * @param is        input stream
     * @param redirect  output stream
     * @param dialog    reference to simulation dialog
     */
    public GCStreamMonitor(InputStream is,
                           OutputStream redirect,
                           SimulationDialog dialog)
    {
        super(is, redirect);
        this.dialog = dialog;
    }

    // =========================================================================
    // PROTECTED INTERFACE
    // =========================================================================

    /**
     * Processes the output from the input stream, checking for embedded
     * messages destined for the GUI. If an embedded message is found, it
     * is formatted appropriately and displayed on the dialog.
     *
     * @param line  string containing external process output
     * @return      string containing possibly reformatted output
     * @throws NullPointerException if line is null
     */
    protected String processLine(String line)
    {
        if (line == null)
            throw new NullPointerException("line");

        // If line doesn't begin with marker, return it unchanged
        if (line.startsWith(KEY_BASE) == false)
            return line;

        StringTokenizer st = new StringTokenizer(line);
        String propertyKey = st.nextToken();

        // Process embedded message
        String outputString = processEmbeddedMessage(propertyKey, st);

        // Update interface
        {
            DialogUpdate dialogUpdate;

            dialogUpdate = new DialogUpdate(outputString,
                                            isPrimaryLabelUpdate(propertyKey),
                                            isProcessing(propertyKey));
            SwingUtilities.invokeLater(dialogUpdate);
        }

        return outputString;
    }

    // =========================================================================
    // PRIVATE INTERFACE
    // =========================================================================

    /**
     * Determines if the property key is for an error message
     *
     * @param propertyKey  string representing resource bundle property key
     * @return             true if error message property key; otherwise, false
     */
    private boolean isErrorMessageKey(String propertyKey)
    {
        return propertyKey.startsWith(ERROR_KEY_BASE);
    }

    /**
     * Determines if the dialog's primary or secondary label should be updated
     *
     * @param propertyKey  string containing resource bundle property key
     * @return             true if primary label update; otherwise, false
     */
    private boolean isPrimaryLabelUpdate(String propertyKey)
    {
        if ((propertyKey.compareTo(MONITOR_APP_PROCESSING_COMPLETE) == 0) ||
            (propertyKey.compareTo(MONITOR_APP_PROCESSING_INPUT) == 0)    ||
            (propertyKey.compareTo(MONITOR_APP_TERMINATED) == 0)          ||
            (propertyKey.compareTo(MONITOR_EXTRACT_CLUSTER) == 0)         ||
            (propertyKey.compareTo(MONITOR_DISPLAY_RESULTS) == 0)         ||
            (propertyKey.compareTo(MONITOR_HIERARCHICAL_CLUSTERING) == 0))
        {
            return true;
        }

        return false;
    }

    /**
     * Determines if this property key indicates the simulation is stopping.
     * Used to indicate the external process should be terminated.
     *
     * @param propertyKey  string containing resource bundle property key
     * @return             true if external process continues; otherwise, false
     */
    private boolean isProcessing(String propertyKey)
    {
        // Determine if user would be stopping early
        if ((propertyKey.compareTo(MONITOR_APP_PROCESSING_COMPLETE) == 0) ||
            (propertyKey.compareTo(MONITOR_APP_TERMINATED) == 0) ||
            (propertyKey.compareTo(MONITOR_ERR_FATAL) == 0))
        {
            return false;
        }

        return true;
    }

    /**
     * Creates an output string from an embedded message.
     * Retrieves message format (using <code>propertyKey</code>) from resource
     * bundle, passes along any arguments, and formats an output string.
     *
     * @param propertyKey  string containing property key
     * @param st           string tokenizer containing embedded message
     * @return             string containing formatted output string
     */
    private String processEmbeddedMessage(String propertyKey,
                                          StringTokenizer st)
    {
        String outputString;

        try
        {
            Object[] args;

            {
                final String defaultDelimiterStr = " \t\n\r\f";
                final String errMsgDelimiterStr = "\n\r\f";
                String delimiterStr = defaultDelimiterStr;

                if (isErrorMessageKey(propertyKey) == true)
                    delimiterStr = errMsgDelimiterStr;

                List vector = new ArrayList();
                while (st.hasMoreTokens() == true)
                    vector.add(st.nextToken(delimiterStr).trim());

                String[] array = null;
                if (vector.size() != 0)
                    array = (String[]) vector.toArray(new String[0]);

                args = array;
            }

            String formatStr = getBundleString(propertyKey);
            MessageFormat mf = new MessageFormat(formatStr);
            outputString = mf.format(args);
        }
        catch (Exception e)
        {
            e.printStackTrace();

            String formatStr = getBundleString(ERR_TRANSLATE_KEY);
            Object[] args = { propertyKey };
            MessageFormat mf = new MessageFormat(formatStr);
            outputString = mf.format(args) + lineSeparator;
        }

        return outputString;
    }

    // =========================================================================
    // INNER CLASSES
    // =========================================================================

    /**
     * DialogUpdate - Updates the simulation dialog status
     * Passed to AWTThread where all GUI updates take place 
     *
     * @version 1.0 2001/08/16
     * @author <a href="mailto:roebuck@mdanderson.org">P. Roebuck</a>
     */
    class DialogUpdate implements Runnable
    {
        // =====================================================================
        // CONSTRUCTOR
        // =====================================================================

        /**
         * Class constructor specifying label text, label destination, and
         * simulation status.
         *
         * @param labelStr      string containing text to display
         * @param isPrimary     if true, update primary label;
         *                      otherwise, update secondary label
         * @param isProcessing  if true, simulation is still running;
         *                      otherwise, simulation has completed
         */
        DialogUpdate(String labelStr, boolean isPrimary, boolean isProcessing)
        {
            this.labelStr = labelStr;
            this.isPrimary = isPrimary;
            this.isProcessing = isProcessing;
        }

        // =====================================================================
        // IMPLEMENT RUNNABLE
        // =====================================================================

        /**
         * Thread entry-point for DialogUpdate object.
         * Causes simulation dialog to be updates appropriately
         *
         * @see SimulationDialog
         */
        public void run()
        {
            if (isPrimary == true)
                dialog.updateExtract(labelStr);
            else
                dialog.updateProgress(labelStr);
            dialog.setProcessing(isProcessing);
        }

        // =====================================================================
        // OBJECT ATTRIBUTES
        // =====================================================================

        private String labelStr;
        private boolean isPrimary = false;
        private boolean isProcessing = false;
    }

    // =========================================================================
    // CLASS INTERFACE
    // =========================================================================

    /**
     * Fetches messages from resource bundle.
     *
     * @param propertyKey  string containing property key for format string
     * @return             string containing message
     */
    static String getBundleString(String propertyKey)
    {
        return (bundle == null) ? null : bundle.getString(propertyKey);
    }

    /**
     * Returns the native platform's line separator
     *
     * @return  string representing line separator
     */
    private static String getLineSeparator()
    {
        String lineSeparator = null;

        try
        {
            lineSeparator = System.getProperty("line.separator");
        }
        catch (Exception e)
        {
            // use newline as default
        }

        if (lineSeparator == null || lineSeparator.length() < 1)
            lineSeparator = "\n";

        return lineSeparator;
    }

    // =========================================================================
    // CONSTANTS
    // =========================================================================

    private static final String KEY_BASE       = "PK_MONITOR_";
    private static final String ERROR_KEY_BASE = KEY_BASE + "ERR_";

    // =========================================================================
    // CLASS ATTRIBUTES
    // =========================================================================

    private static ResourceBundle bundle;
    private static String lineSeparator;

    static
    {
        final String propertyFile = GCStreamMonitor.class.getName();
        bundle = ResourceBundle.getBundle(propertyFile);

        lineSeparator = getLineSeparator();
    }

    // =========================================================================
    // OBJECT ATTRIBUTES
    // =========================================================================

    private SimulationDialog dialog;
}

