/*=================== FILE ABSTRACT ==========================================
 *
 * Synopsis::         Source for GeneClust user interface (SimulationDialog)
 *
 * File Name::        %M%
 * Version::          %I%
 * Type::             source
 * Date Completed::   %G%
 *
 * Environment::
 *     Language::     Java
 *     Computer::     Compaq/Alpha
 *     OS::           DigitalUNIX
 *     OSVersion::    4.0
 *
 * Module Information::
 *     Owner::        MDACC
 *     Department::   Biostatistics
 *     Researcher::   Kim-Anh Do
 *     Application::  GeneClust
 *
 *============================================================================*/

/*
 * @(#)SimulationDialog.java    0.95 2000/02/05
 *
 * Copyright (c) 2000 M.D. Anderson Cancer Center
 * 1515 Holcombe, Houston, TX, 77030, U.S.A.
 * All rights reserved.
 */

package org.mdanderson.biostat.geneclust;

import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.Point;
import javax.swing.JDialog;
import javax.swing.JOptionPane;


/**
 * SimulationDialog - UI bridge that displays simulation output in dialog
 *
 * @version 0.90 2001/02/05
 * @author <a href="mailto:roebuck@mdanderson.org">P. Roebuck</a>
 */
public class SimulationDialog
    implements GeneClustKeys
{
    // =========================================================================
    // CONSTRUCTOR
    // =========================================================================

    /**
     * Class constructor specifying parent.
     *
     * @param parent  container object
     */
    public SimulationDialog(Container parent)
    {
        final String title = getBundleString(SIMULATION_DIALOG_TITLE);
        final String buttonLabel =
                         getBundleString(SIMULATION_DIALOG_BUTTON_LABEL);

        extractMessage = getBundleString(SIMULATION_DIALOG_MESSAGE);
        Object[] messages = { extractMessage, progressMessage };
        Object[] buttonLabels = { buttonLabel };
        simulationPane = new JOptionPane(messages,
                                         JOptionPane.PLAIN_MESSAGE,
                                         JOptionPane.DEFAULT_OPTION,
                                         null,
                                         buttonLabels);

        // Create modal dialog!
        dialog = simulationPane.createDialog(parent, title);
        dialog.setResizable(true);

        Insets insets = dialog.getInsets();
        final int WIDTH  = 380;
        final int HEIGHT = 125;

        int width  = WIDTH  + (insets.left + insets.right);
        int height = HEIGHT + (insets.top + insets.bottom);

        dialog.setSize(width, height);
        dialog.validate();
    }

    // =========================================================================
    // PUBLIC INTERFACE
    // =========================================================================

    /**
     * Displays this model dialog.
     */
    public void displayAndWait()
    {
        centerOnParent();
        dialog.setVisible(true);
        // With modal property in place, code waits here until OK pressed...
    }

    /**
     * Returns the state of the simulation.
     *
     * @return  true if simulation is still processing;
     *          otherwise, it has stopped.
     * @see #setProcessing
     */
    public boolean getProcessing()
    {
        return isProcessing;
    }

    /**
     * Sets the state of the simulation.
     *
     * @param isProcessing  if true, simulation is still processing;
     *                      otherwise, it has stopped.
     * @see #getProcessing
     */
    public void setProcessing(boolean isProcessing)
    {
        this.isProcessing = isProcessing;
    }

    /**
     * Updates the primary message label for simulation progress and blanks
     * the secondary message label.
     *
     * @param extractMessage  string containing primary message label text
     */
    public synchronized void updateExtract(final String extractMessage)
    {
        this.extractMessage = extractMessage;
        this.progressMessage = BLANK;
        updateLabels();
    }

    /**
     * Updates the secondary message label for simulation progress.
     *
     * @param progressMessage  string containing secondary message label text
     */
    public synchronized void updateProgress(final String progressMessage)
    {
        this.progressMessage = progressMessage;
        updateLabels();
    }

    // =========================================================================
    // PROTECTED INTERFACE
    // =========================================================================

    /**
     * Centers the dialog on its parent
     */
    protected void centerOnParent()
    {
        Container parent = dialog.getParent();
        Point topLeft = parent.getLocationOnScreen();

        Dimension parentSize = parent.getSize();
        Dimension size = dialog.getSize();

        int x = topLeft.x;
        int y = topLeft.y;

        if (parentSize.width > size.width)
            x += ((parentSize.width - size.width) / 2);

        if (parentSize.height > size.height)
            y += ((parentSize.height - size.height) / 2);

        dialog.setLocation(x, y);
    }

    // =========================================================================
    // PRIVATE INTERFACE
    // =========================================================================

    /**
     * Updates the dialog message labels with current label text
     */
    private void updateLabels()
    {
        Object[] messages = {
            extractMessage,
            progressMessage
        };
        simulationPane.setMessage(messages);
    }

    // =========================================================================
    // CLASS INTERFACE
    // =========================================================================

    /**
     * Fetches messages from resource bundle.
     *
     * @param propertyKey  string containing property key for format string
     * @return             string containing message
     */
    static String getBundleString(String propertyKey)
    {
        return GeneClust.getBundleString(propertyKey);
    }

    // =========================================================================
    // CLASS ATTRIBUTES
    // =========================================================================

    private static final String BLANK;

    static
    {
        BLANK = "                                                  ";
    }

    // =========================================================================
    // OBJECT ATTRIBUTES
    // =========================================================================

    private JDialog dialog;
    private JOptionPane simulationPane;
    private String extractMessage = BLANK;
    private String progressMessage = BLANK;
    private boolean isProcessing = true;
}

