/*=================== FILE ABSTRACT ==========================================
 *
 * Synopsis::         Source for validated numeric textfield class
 *
 * File Name::        %M%
 * Version::          %I%
 * Type::             source
 * Date Completed::   %G%
 *
 * Environment::
 *     Language::     Java
 *     Computer::     Compaq/Alpha
 *     OS::           DigitalUNIX
 *     OSVersion::    4.0
 *
 * Module Information::
 *     Owner::        MDACC
 *     Department::   Biostatistics
 *     Researcher::   N/A
 *     Application::  <generic>
 *
 *============================================================================*/

/*
 * @(#)NumericTextField.java    1.0 2000/12/26
 *
 * Copyright (c) 2000 M.D. Anderson Cancer Center
 * 1515 Holcombe, Houston, TX, 77030, U.S.A.
 * All rights reserved.
 */

package org.mdanderson.biostat.util;

import javax.swing.JTextField;
import javax.swing.plaf.metal.MetalLookAndFeel;


/**
 * NumericTextField - Allows numeric validation of a textfield.
 *
 * @version 1.0 2000/12/26
 * @author <a href="mailto:roebuck@mdanderson.org">P. Roebuck</a>
 */
public class NumericTextField extends JTextField
  implements StatusListener
{
    // =========================================================================
    // CONSTRUCTORS
    // =========================================================================

    /**
     * Class constructor
     */
    public NumericTextField()
    {
        super();
        initValidation();
        setHorizontalAlignment(JTextField.RIGHT);

    }

    /**
     * Class constructor specifying text
     *
     * @param text  string containing initial value
     */
    public NumericTextField(String text)
    {
        super(text);
        initValidation();
    }

    /**
     * Class constructor specifying number of columns
     *
     * @param columns  number of columns to use to calculate the preferred width
     * @throws IllegalArgumentException if columns argument is negative
     */
    public NumericTextField(int columns)
    {
        super(columns);
        initValidation();
    }

    /**
     * Class constructor specifying text and number of columns
     *
     * @param text     string containing initial value
     * @param columns  number of columns to use to calculate the preferred width
     * @throws IllegalArgumentException if columns argument is negative
     */
    public NumericTextField(String text, int columns)
    {
        super(text, columns);
        initValidation();
    }

    // =========================================================================
    // PUBLIC INTERFACE
    // =========================================================================

    /**
     * Sets its numeric validator's allowed range of values
     *
     * @param min  minimum double value allowed (inclusive)
     * @param max  maximum double value allowed (inclusive)
     * @see NumericValidator#setRange
     */
    public void setRange(double min, double max)
    {
        nv.setRange(min, max);
    }

    /**
     * Adds its numeric validator as a status listener for
     * notification of state of data changes
     *
     * @param listener  status listener
     * @see ValidationAdapter#addStatusListener
     */
    public void addStatusListener(StatusListener listener)
    {
        nv.addStatusListener(listener);
    }

    // =========================================================================
    // IMPLEMENTS STATUSLISTENER
    // =========================================================================

    /**
     * Notifies observers when the state of the data changes.
     * Swaps the foreground and background colors depending on data validity.
     *
     * @param event  status event
     */
    public void statusChanged(StatusEvent event)
    {
        if (event.getStatus())
        {
            setBackground(MetalLookAndFeel.getWhite());
            setForeground(MetalLookAndFeel.getBlack());
            setCaretColor(MetalLookAndFeel.getBlack());
        }
        else
        {
            setBackground(MetalLookAndFeel.getBlack());
            setForeground(MetalLookAndFeel.getWhite());
            setCaretColor(MetalLookAndFeel.getWhite());
        }
    }

    // =========================================================================
    // PACKAGE INTERFACE
    // =========================================================================

    /**
     * Initializes the validation features for this textfield
     */
    void initValidation()
    {
        // Keystroke filtering
        filter = new NumericFilter(this);

        // Asynchronous validation
        nv = new NumericValidator();
        addStatusListener(this);
        getDocument().addDocumentListener(nv);
    }

    // =========================================================================
    // OBJECT ATTRIBUTES
    // =========================================================================

    Filter filter;
    NumericValidator nv;
}

