/*=================== FILE ABSTRACT ==========================================
 *
 * Synopsis::         Source for gene clustering routines
 *
 * File Name::        %M%
 * Version::          %I%
 * Type::             source
 * Date Completed::   %G%
 *
 * Environment::
 *     Language::     ANSI C
 *     Computer::     Compaq/Alpha
 *     OS::           DigitalUNIX
 *     OSVersion::    4.0
 *
 * Module Information::
 *     Owner::        MDACC
 *     Department::   Biostatistics
 *     Researcher::   Kim-Anh Do
 *     Application::  geneclust
 *
 *============================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <errno.h>
#include <assert.h>
#include "_bool.h"
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include "Utility.h"
#include "GeneClust.h"
#include "LinkedList.h"


#define TRACE_FNS

/*
 * Defines
 */
#define RECOMPUTE_M_PRECISION               1.0e-3
#define SHAVING_THRESHOLD	            20
#define GENE_LABELS_PREFIX                  "GENE"
#define SAMPLE_LABELS_PREFIX                "sample"
#define CLUSTER_LABELS_PREFIX               "clabels"
#define EIGENVALUE_FILE_PREFIX              "eigen"
#define GAP_FILE_PREFIX                     "gap"
#define PRINCIPAL_COMPONENT_FILE_PREFIX     "pc"
#define VARIANCE_FILE_PREFIX                "variance"
#define FILE_EXTENSION_SUFFIX               ".txt"


/*
 * Datatype Defines
 */
#define MAX_MESSAGE_LEN                     63


/*
 * Macros
 */
#define BEEP()                              printf("\007")
#define PAUSE()                             sleep(3)
#define PROCESS_VAR_ARGS(message,format,args)   \
    do {                                        \
        va_start (args, format);                \
        (void)vsprintf (message, format, args); \
        va_end (args);                          \
    } while (0)


/*
 * Typedefs
 */
typedef char Message_t[MAX_MESSAGE_LEN+1];

typedef struct Perm {
    unsigned index;
    int val;
} Perm_t;

typedef struct Result {
    /* N == total number of genes == clusterSize[0]                           */
    /* clusterSize[numShavings] == 1, to simplify later code.                 */
    int      numShavings;           /* Number of shavings in this cluster     */
    ivector_t *clusterSize;         /* Number of genes in each cluster        */
                                    /*   (vector[0..numShavings])             */
    ivector_t *geneMap;             /* Genes in cluster order (vector[0..N-1])*/
    unsigned optIndex;              /* Index of optimal cluster               */
    double   eigenvalue;            /* Eigenvalue of optimal cluster          */
    vector_t *principalComponent;   /* Principal component of optimal cluster */
    size_t   principalComponentSize;/* Length of principal component          */
    vector_t *weight;               /* Vector [0..N-1] giving influence of    */
                                    /*   each gene                            */
} Result_t;

    /*
     * Additional notes on geneMap vector
     * ==================================
     *
     * The genes (row numbers) in each cluster are stored in the vector geneMap.
     * Since the genes in each successive cluster are a subset of those in
     * earlier clusters, a single vector suffices. Specifically, the genes in
     * cluster i are stored in geneMap[0..clusterSize[i]-1].
     *
     * There is no ordering of genes within each cluster (they are sets). In
     * fact, during construction, the order of genes within an already
     * determined cluster changes as subclusters are identified.
     */


/*
 * Module Variables
 */
static bool gShowProgress = false;


/*
 * Private Routine Declarations
 */
static bool AreMoreClustersLeft (int);
static void CalcAvgSignedGene (matrix_t *, int, int, vector_t *, ivector_t *, vector_t *);
static Result_t *CloneResult (int, int, const Result_t *);
static int ComparePerm (const void *, const void *);
static matrix_t *ComputeQ (int, const char *, int *);
static void ComputeXStar (matrix_t *, matrix_t *, int, int, matrix_t *, int);
static void ComputeXtransposeX (matrix_t *, matrix_t *, int, int, const ivector_t *);
static void ConstructLabels (char **, int, const char *, const char *);
static void ConstructQColumn (matrix_t *, unsigned, unsigned, const unsigned *);
static void DisplayProgress (const char *, ...);
static void DoNestedShave (Result_t *, matrix_t *, matrix_t *, int, int,
                           vector_t **, matrix_t *, vector_t *);
static void DotProd (ivector_t *, matrix_t *, int, int, vector_t *, vector_t *, vector_t *);
static void EstimateOptimal (Result_t *, matrix_t *, int, int, int, double,
                             int, int, matrix_t *, int);
static void FindCluster (matrix_t *, matrix_t *, int, int, double, int, double,
                         int, int, char **, char **, char **, matrix_t *, int, 
                         Result_t *);
static void FindPC (matrix_t *, int, vector_t *, double *);
static void FindR2 (Result_t *, vector_t *, matrix_t *, int, int, matrix_t *,
                    vector_t *, vector_t *, vector_t *);
static void FreeResult (Result_t *, int, int);
static bool IsDirectory (const char *);
static bool IsRegularFile (const char *);
static Result_t *NewResult (int, int, double);
static void Orthogonalize (matrix_t *, vector_t *, int, int);
static void PermuteMatrixRowElements (matrix_t *, matrix_t *, int, int);
static int ShaveCount (int, double);
static void ShaveOff (vector_t *, ivector_t *, int, int);

static void StoreClusterInfo (matrix_t *, int, int, const ivector_t *,
                              char **, char **, char **, const vector_t *, int);
static void StoreEigenvalue (int, double);
static void StoreGapInfo (int, int, int, const ivector_t *, const vector_t *, const matrix_t *);
static void StoreLabels (char **, int, const ivector_t *, const char *);
static void StoreMatrix (matrix_t *, int, int);
static void StorePrincipalComponent (int, size_t, const vector_t *);
static void StoreVarianceInfo (int, int, int, const ivector_t *,
                               const vector_t *, const vector_t *,
                               const vector_t *, const vector_t *, const vector_t *,
			       const matrix_t *);
static void UpdateXtransposeX (matrix_t *, matrix_t *, int, int, const ivector_t *);


/*
 * Public Routines
 */

/*============================================================================
 *
 * Name:: GeneShave
 *
 * Description::
 *     Performs nested shaving and stores the Gene labels, sample labels
 *     and cluster lengths in separate files to be opened with S-plus.
 *
 * Inputs::
 *     outputDirectoryName - Directory into which the output will be stored.
 *     numClusters     - Number of clusters to find.
 *     percentToShave  - Percent to shave.
 *     numPermutations - Number of permutations to find optimal cluster.
 *     gapTolerance    - Gap tolerance.
 *     V               - Vector representing gene expr matrix in row-major order
 *     row             - Pointer to number of genes (rows in gene expr matrix).
 *     col             - Pointer to number of samples (cell lines).
 *     verbose         - Show progress of computation?
 *     generateLabels  - Determines whether labels created or read from file.
 *     writeMatrix     - Determines whether to write matrix to file (demo mode).
 *     percentSupervision - Supervised shaving percentage (0.00 - 1.00).
 *     classificationFileName - Classification file required for supervised shaving.
 *
 * Outputs::
 *     numClustersExtracted - Actual number of clusters extracted.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

void GeneShave (
  SCharacter_t *outputDirectoryName,
  SInteger_t *numClusters,
  SInteger_t *percentToShave,
  SInteger_t *numPermutations,
  SInteger_t *gapTolerance,
  SDouble_t  *V,
  SInteger_t *row,
  SInteger_t *col,
  SLogical_t *verbose,
  SLogical_t *generateLabels,
  SLogical_t *writeMatrix,
  SInteger_t  *percentSupervision,
  SCharacter_t *classificationFileName,
  SInteger_t *numClustersExtracted)
{
    matrix_t *X;
    matrix_t *originalX;
    int N;
    int p;
    int NCLST;
    matrix_t *Q = NULL;
    int numClasses;
    char **sglab = NULL;     /* String arrays of gene labels */
    char **sslab = NULL;     /* String arrays of sample labels */
    char **point = NULL;     /* String arrays of cluster filenames */
    char **clbl  = NULL;     /* String arrays of cluster label filenames */
    PathName_t rowPathName;
    PathName_t colPathName;

    FILE *clusterLengthsFile;

    assert (outputDirectoryName != NULL);
    assert (numClusters != NULL);
    assert (percentToShave != NULL);
    assert (numPermutations != NULL);
    assert (gapTolerance != NULL);
    assert (V != NULL);
    assert (row != NULL);
    assert (col != NULL);
    assert (verbose != NULL);
    assert (generateLabels != NULL);
    assert (writeMatrix != NULL);
    assert (percentSupervision != NULL);
    assert (*percentSupervision >= 0 && *percentSupervision <= 100);
    assert (numClustersExtracted != NULL);

#ifdef TRACEFNS
    fprintf (stderr, "> GeneShave\n");
#endif
    SetOutputDirectory ((char *)*outputDirectoryName);
    N = (int)*row;              /* Number of genes */
    p = (int)*col;              /* Number of cell boundaries */
    NCLST = (int)*numClusters;  /* Number of clusters */

    gShowProgress = ((int)*verbose) ? true : false;

    X = dmatrix (0, N-1, 0, p-1);

    sglab = cvector (N);
    sslab = cvector (p);
    point = cvector (NCLST);
    clbl  = cvector (NCLST);

#define DEBUG
#ifdef DEBUG
    fprintf (stderr, "*** Write Matrix: %s\n",
             (bool)*writeMatrix ? "True" : "False");
    switch((int)*percentSupervision)
    {
        case 0:
            fprintf (stderr, "*** No Supervision\n");
            break;

        case 100:
            fprintf (stderr, "*** Full Supervision\n");
            break;

        default:
            fprintf (stderr, "*** Partial Supervision (%2d%%)\n",
                     (int)*percentSupervision);
            break;
    }
    fprintf (stderr, "*** Gap Tolerance = %d\n", (int)*gapTolerance);
#endif

    /* If supervised, input classification file and construct Q matrix */
    if (*percentSupervision != 0)
    {
        bool isPartialSupervision = (((int)*percentSupervision) != 100) ? true :
                                                                          false;
        assert (classificationFileName != NULL);

#ifdef DEBUG
fprintf (stderr, "*** ClassFileName: %s\n", (char *)*classificationFileName);
#endif

        if (IsRegularFile (*classificationFileName) == false)
        {
            ErrMessage_t errMessage;

            sprintf (errMessage, "GeneShave(\"%s\"): %s\n",
                     *classificationFileName,
                     "not a regular file");
            Abend (errMessage);
            /*NOTREACHED*/
        }

        Q = ComputeQ (p, (const char *) *classificationFileName, &numClasses);

        /*
         * If partial supervision is requested, we compute the partial
         * supervision operator Hy = (1-supervision)*I + supervision*Q*Q^T
         * The result is a positive definite matrix, so we can use the
         * Cholesky decomposition to compute its "square-root" A, such that
         * A*A^T = Hy. We then substitute this matrix for Q. The
         * remaining computational details are identical to the fully
         * supervised case.
         */
        if (isPartialSupervision)
        {
            register int i;
            double supervision = ((int)*percentSupervision) / 100.0;
            vector_t *diag = dvector (0, p-1);             /* Used 1 origin for NR */
            matrix_t *Hy = dmatrix (0, p-1, 0, p-1);        /* Used 1 origin for NR */
            matrix_t *A = dmatrix (0, p-1, 0, p-1);

            /* Compute Hy = (1.0 - supervision) * I + supervision * QQ^T */
            for (i = 0; i < p; i++)
            {
                register int j;

                for (j = 0; j < p; j++)
                {
                    register int k;
                    double total = 0.0;

                    for (k = 0; k < numClasses; k++)
                        total += gsl_matrix_get (Q, i, k) * gsl_matrix_get (Q, j, k);
                    gsl_matrix_set (Hy, i, j, total * supervision);
                }
                gsl_matrix_set (Hy, i, i,
		             gsl_matrix_get (Hy, i, i) + 1.0 - supervision);
            }

            /* Compute cholesky decomposition of Hy */
            /* choldc(Hy, p, diag); */
	    gsl_linalg_cholesky_decomp (Hy);

            /* Copy and expand packed decomposition of Hy into A */
            for (i = 0; i < p; i++)
            {
                register int j;

                for (j = 0; j < i; j++)
                    gsl_matrix_set (A, i, j, gsl_matrix_get (Hy, i, j));
                gsl_matrix_set (A, i, i, gsl_matrix_get (Hy, i, i));
                for (j = i+1; j < p; j++)
                    gsl_matrix_set (A, i, j, 0.0);
            }

            /* Replace Q by A */
            free_dmatrix (Q, 0, p-1, 0, numClasses-1);
            numClasses = p;             /* Number of columns in Hy */
            Q = A;

            /* Release memory */
            free_dvector (diag, 0, p-1);
            free_dmatrix (Hy, 0, p-1, 0, p-1);
        }
    }

    /* Unfold matrix */
    FetchMatrixAsRowMajor (V, X, N, p);
    CenterData (X, N, p);

    /* Create copy of original matrix */
    originalX = dmatrix (0, N-1, 0, p-1);
    gsl_matrix_memcpy (originalX, X);
#ifdef NR
    {
        register unsigned i;

        for (i = 0; i < N; i++)
        {
            register unsigned j;

            for (j = 0; j < p; j++)
                originalX[i][j] = X[i][j];
        }
    }
#endif

    /* Write a copy of the original matrix to disk */
    if ((bool) *writeMatrix)
        StoreMatrix (X, N, p);

    /* Initialize file names accordingly */
    sprintf (rowPathName, "%s/%s", GetOutputDirectory (), FILE_ROW_LABELS);
    sprintf (colPathName, "%s/%s", GetOutputDirectory (), FILE_COL_LABELS);

    if (*generateLabels)
    {
        /* Create the labels for the samples and store them */
        ConstructLabels (sglab, N, GENE_LABELS_PREFIX, "");

        StoreLabels (sglab, N, NULL, rowPathName);

        ConstructLabels (sslab, p, SAMPLE_LABELS_PREFIX, "");
        StoreLabels (sslab, p, NULL, colPathName);
    }
    else
    {
        FILE *rowFile;           /* Row labels file */
        FILE *colFile;           /* Column labels file */
        Buffer_t buffer;
        register int i;

        /* Open input files */
        if ((rowFile = fopen (rowPathName, "r")) == NULL)
        {
            ErrMessage_t errMessage;

            sprintf (errMessage, "GeneShave(\"%s\"): %s\n",
                     rowPathName,
                     strerror (errno));
            Abend (errMessage);
            /*NOTREACHED*/
        }

        if ((colFile = fopen (colPathName, "r")) == NULL)
        {
            ErrMessage_t errMessage;

            sprintf (errMessage, "GeneShave(\"%s\"): %s\n",
                     colPathName,
                     strerror (errno));
            Abend (errMessage);
            /*NOTREACHED*/
        }

        /* Read strings from input files */
        for (i = 0; i < N; i++)
        {
            fscanf (rowFile, "%s\n", buffer);
            strcpy (sglab[i], buffer);
        }
        for (i = 0; i < p; i++)
        {
            fscanf (colFile, "%s\n", buffer);
            strcpy (sslab[i], buffer);
        }

        /* Close input files */
        fclose (rowFile);
        fclose (colFile);
    }

    {
        PathName_t clusterLabelsPathName;

        sprintf (clusterLabelsPathName, "%s/%s",
                 GetOutputDirectory (), FILE_CLUSTER_LABELS);
        ConstructLabels (clbl,
                         NCLST,
                         CLUSTER_LABELS_PREFIX,
                         FILE_EXTENSION_SUFFIX);
        StoreLabels (clbl, NCLST, NULL, clusterLabelsPathName);
    }

    {
        PathName_t clustersPathName;

        sprintf (clustersPathName, "%s/%s",
                 GetOutputDirectory (), FILE_CLUSTERS);
        ConstructLabels (point, NCLST, "cluster", FILE_EXTENSION_SUFFIX);
        StoreLabels (point, NCLST, NULL, clustersPathName);
    }

    /* Open output file */
    {
        PathName_t clusterLengthsPathName;

        sprintf (clusterLengthsPathName, "%s/%s",
                 GetOutputDirectory (), FILE_CLUSTER_LENGTHS);
        if ((clusterLengthsFile = fopen (clusterLengthsPathName, "w")) == NULL)
        {
            ErrMessage_t errMessage;

            sprintf (errMessage, "GeneShave(\"%s\"): %s\n",
                     clusterLengthsPathName,
                     strerror (errno));
            Abend (errMessage);
            /*NOTREACHED*/
        }
    }

    /* Compute */
    {
        register int i;
        double percentage = ((int) *percentToShave) / 100.0;
        int nperm = ((int) *numPermutations);
        double tolerance = (double)((int) *gapTolerance);
        Result_t *result = NewResult (N, p, percentage);

#ifdef DEBUG
        fprintf (stderr, "  ShavePercent = %f\n", percentage);
        fprintf (stderr, "  Number of Shavings/cluster : %d\n",
                 result->numShavings);
#endif

        /* Loop on number of clusters to find */
        for (i = 1; i <= NCLST; i++)
        {
            DisplayProgress ("%s %d", EXTRACT_CLUSTER, i);

            FindCluster (X, originalX, N, p, percentage, nperm, tolerance,
                         i, NCLST, sglab, point, clbl,
                         Q, numClasses, result);

            /* Write output */
            fprintf (clusterLengthsFile, "%d\t %d\n",
                     result->optIndex+1,
		     gsl_vector_int_get (result->clusterSize, result->optIndex));
            StoreEigenvalue (i, result->eigenvalue);
            StorePrincipalComponent (i, p, result->principalComponent);

#if 0 && defined(DEBUG)
            fprintf (stderr, "num[%d]= %d\t optIndex=%d eigenvalue=%f\n",
                     i-1, result->clusterSize[result->optIndex],
                     result->optIndex, result->eigenvalue);
            {
                register unsigned j;

                fprintf (stderr, "principal component=[");
                for (j = 0; j < p; j++)
                    fprintf (stderr, "%f ", result->principalComponent[j]);
                fprintf (stderr, "]\n");
            }
#endif

            *numClustersExtracted = i;
            if (i < NCLST)
            {
                if (AreMoreClustersLeft (result->optIndex) == false)
                {
                    DisplayProgress (NO_CLUSTERS_LEFT);
                    fprintf (stderr, "No clusters left to extract...\n");
                    fprintf (stderr, "\tExtracted %d of requested %d\n",
                             i, NCLST);
                    /* Alert the user the problem occurred */
                    BEEP();
                    PAUSE();
                    break;
                }
            }
        }
#if 0 && defined(DEBUG)
        fprintf (stderr, "#Clusters Extracted: %ld\n", *numClustersExtracted);
#endif
        FreeResult (result, N, p);
        result = NULL;
    }

    /* Close outfile */
    fclose (clusterLengthsFile);

    /* Free storage  */
#ifdef LABELS
    free_cvector (sglab, N);
    free_cvector (sslab, p);
    free_cvector (clbl, NCLST);
#endif
    free_cvector (point, NCLST);
    free_dmatrix (X, 0, N-1, 0, p-1);
    free_dmatrix (originalX, 0, N-1, 0, p-1);
    if (Q != NULL)
        free_dmatrix (Q, 0, p-1, 0, numClasses-1);

#ifdef TRACEFNS
    fprintf (stderr, "< GeneShave\n");
#endif
}


/*============================================================================
 *
 * Name:: GetOutputDirectory
 *
 * Description::
 *     Determines the location where output files should be written. Checks
 *     environment variable for directory path. If the environment variable
 *     is defined and its value is an existing directory, then its value
 *     is used as the output directory; otherwise, an internal default is
 *     returned.
 *
 * Inputs::
 *     (None)
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     Returns string containing the pathname for the output directory.
 *
 *============================================================================*/

static const char *outputDirectory = RELPATH_OUTPUT_DIRECTORY;

void
SetOutputDirectory (const char *dirName)
{
    /* Make sure value is a directory */
    if (IsDirectory (dirName) == true)
	outputDirectory = dirName;
    else
    {
	/* Warn the user of the problem */
	fprintf (stderr,
		 "Output path \"%s\" does not refer to\
		 a directory. Using default path <%s> instead.",
		 ENVVAR_GCOUTPUT, outputDirectory);
	BEEP();
    }
}

const char *GetOutputDirectory (void)
{
    return (outputDirectory);
}


/*============================================================================
 *
 * Name:: Abend
 *
 * Description::
 *     Writes error message and terminates the application.
 *
 * Inputs::
 *     message - Error message to display before terminating.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     Returns exit status of the application to the OS via exit().
 *
 *============================================================================*/

void Abend (const char *message)
{
    assert (message != NULL);

    if (gShowProgress == true)
    {
        DisplayProgress ("%s", ABEND);
        DisplayProgress ("%s %s", REPORT_ERROR, message);
    }
    else
    {
        fprintf (stdout, "%s\n", message);
        fflush (stdout);
    }

    exit (EXIT_FAILURE);
    /*NOTREACHED*/
}


/*
 * Private Routines
 */

/*============================================================================
 *
 * Name:: AreMoreClustersLeft
 *
 * Description::
 *     Determines whether there are any unprocessed clusters left in matrix.
 *
 * Inputs::
 *     optimalIndex - Index of the optimal cluster in the matrix.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     Returns true if more clusters left to process; otherwise, false.
 *
 *============================================================================*/

static bool AreMoreClustersLeft (int optimalIndex)
{
    return ((optimalIndex != 0) ? true : false);
}


/*============================================================================
 *
 * Name:: CalcAvgSignedGene
 *
 * Description::
 *     Finds the average signed gene per cluster.
 *
 * Inputs::
 *     X           - Gene expression matrix.
 *     p           - Number of columns (samples/cells) in X.
 *     clusterSize - Cluster size.
 *     idx         - Vector [0..clusterSize-1] storing the row numbers
 *                   of the genes in the cluster.
 *     weight      - Vector [genes] storing the weights of the PCs
 *                   for the genes in the cluster.
 *
 * Outputs::
 *     (None)
 *
 * Inputs / Outputs::
 *     avgGene     - Average signed gene for cluster (vector [0..p-1]).
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void CalcAvgSignedGene (
  matrix_t *X,
  int p,
  int clusterSize,
  vector_t *avgGene,
  ivector_t *idx,
  vector_t *weight)
{
    register int i;
    register int j;

#ifdef TRACEFNS
    fprintf (stderr, "> CalcAvgSignedGene\n");
#endif
    assert (X != NULL);
    assert (p > 0);
    assert (clusterSize > 0);
    assert (avgGene != NULL);
    assert (idx != NULL);
    assert (weight != NULL);

    for (j = 0; j < p; j++)
        gsl_vector_set (avgGene, j, 0.0);

    for (i = 0; i < clusterSize; i++)
    {
        int gene = gsl_vector_int_get (idx, i);

        if (gsl_vector_get (weight, gene) > 0) {
            for (j = 0; j < p; j++) {
	        double amount = gsl_matrix_get (X, gene, j);
                gsl_vector_set (avgGene, j,
	                        gsl_vector_get (avgGene, j) + amount);
	    }
	}
	else {
            for (j = 0; j < p; j++) {
	        double amount = gsl_matrix_get (X, gene, j);
                gsl_vector_set (avgGene, j,
	                        gsl_vector_get (avgGene, j) - amount);
	    }
	}
    }

    for (j = 0; j < p; j++) {
        gsl_vector_set (avgGene, j, 
	    gsl_vector_get (avgGene, j) / clusterSize);
    }
#ifdef DEBUG
    fprintf (stderr, "  AvgGene =");
    for (j = 0; j < p; j++) {
        fprintf (stderr, " %g", gsl_vector_get (avgGene, j));
    }
    fprintf (stderr, "\n");
#endif
#ifdef TRACEFNS
    fprintf (stderr, "< CalcAvgSignedGene\n");
#endif
}


/*============================================================================
 *
 * Name:: CloneResult
 *
 * Description::
 *     Allocates a new result structure whose component sizes are the same
 *     as those in the cluster passed as a parameter.
 *
 * Inputs::
 *     N   - Number of genes.
 *     p   - Number of samples per gene.
 *     old - Result structure to clone.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     Returns pointer to newly created instance of Result_t.
 *
 *============================================================================*/

static Result_t *CloneResult (int N, int p, const Result_t *old)
{
    Result_t *result;
    register int i;

    assert (N > 0);
    assert (p > 0);
    assert (old != NULL);

    result = (Result_t *) calloc (1, sizeof(Result_t));
    if (result == NULL)
    {
         ErrMessage_t errMessage;

         sprintf (errMessage, "Error creating result structure: %s\n",
                  strerror (errno));
         Abend (errMessage);
         /*NOTREACHED*/
    }

    result->numShavings = old->numShavings;
    result->clusterSize = ivector (0, result->numShavings);
    gsl_vector_int_memcpy (result->clusterSize, old->clusterSize);

    result->geneMap = ivector (0, N-1);
    result->weight = dvector (0, N-1);
    result->principalComponent = dvector (0, p-1);

    return result;
}


/*============================================================================
 *
 * Name:: ComparePerm
 *
 * Description::
 *     Used to compare two Perm values and determine whether the first
 *     element in the comparison is considered less than, equal to, or
 *     greater than the second element (like strcmp()).
 *
 * Inputs::
 *     pv1 - Pointer to first Perm_t.
 *     pv2 - Pointer to second Perm_t.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     Returns 1 if second element is larger, -1 if second element is smaller,
 *     or 0 in the unlikely event that both elements are the same.
 *
 *============================================================================*/

static int ComparePerm (const void *pv1, const void *pv2)
{
    const Perm_t *ps1 = (const Perm_t *)pv1;
    const Perm_t *ps2 = (const Perm_t *)pv2;

    assert (ps1 != NULL);
    assert (ps2 != NULL);

    if (ps1->val > ps2->val)
        return 1;
    else if (ps1->val < ps2->val)
        return -1;
    else
        return 0;
}


/*============================================================================
 *
 * Name:: ComputeQ
 *
 * Description::
 *    Calculates the Q matrix.
 *    (Q with dim P x |C| p.30 of T. Hastie paper)
 *
 * Inputs::
 *     p             - Number of columns (samples/cells) in gene expr matrix.
 *     classFileName - Column classification file name.
 *
 * Outputs::
 *     numClasses    - |C|
 *
 * Return Value::
 *     Returns newly allocated matrix whose rows are standardized versions
 *     of the class means for each gene.
 *
 *============================================================================*/

static matrix_t *ComputeQ (
  int p,
  const char *classFileName,
  int *numClasses)
{
    matrix_t *Q = NULL;
    FILE *classFile;
    Buffer_t buffer;
    Node_t *list = NULL;
    int numberOfClasses = 0;
    unsigned lineCounter = 0;

    assert (classFileName != NULL);
    assert (numClasses != NULL);

    /* Open input file */
    if ((classFile = fopen (classFileName, "r")) == NULL)
    {
        ErrMessage_t errMessage;

        sprintf (errMessage, "ComputeQ(\"%s\"): %s\n",
                 classFileName,
                 strerror (errno));
        Abend (errMessage);
        /*NOTREACHED*/
    }

    /* Initialize list */
    ListInit (&list);

    /* Calculate number of classes and number of samples in each class */
    while (fgets (buffer, MAX_BUFFER_LEN, classFile) != NULL)
    {
        /* Remove trailing newline characters */
        {
            char *ptr;

            if ((ptr = strrchr (buffer, '\n')) != NULL)
                *ptr = '\0';
        }

        if (strlen (buffer) != 0)
        {
#if 0 && defined(DEBUG)
            fprintf (stderr, "[%02d] buffer=[%s]  val = %d\n",
                     lineCounter, buffer, atoi (buffer));
#endif
            /* Store inputs into a linked list */
            ListAppend (&list, atoi (buffer), lineCounter++);
        }
    }
    fclose (classFile);
    numberOfClasses = ListLength (list);
#if 1 && defined(DEBUG)
    ListPrint (list);
#endif

    /* Quick error check */
    if (lineCounter != p)
    {
        ErrMessage_t errMessage;

        sprintf (errMessage,
                 "Classification file line count(%d) must equal columns(%d)\n",
                 lineCounter, p);

        Abend (errMessage);
        /*NOTREACHED*/
    }

    /* Construct the Q matrix */
    Q = dmatrix (0, p-1, 0, numberOfClasses-1);
    gsl_matrix_set_zero (Q);

    /* Traverse the linked list and construct the Q matrix */
    {
        register Node_t *node;
        unsigned column = 0;

        for (node = list; node != NULL; node = node->next)
        {
            ConstructQColumn (Q, column++, node->numPositions, node->positions);
        }
    }

#if 1 && defined(DEBUG)
    { int i,j;
    /* Print out Q */
    for (i = 0; i < p; i++)
    {
        for (j = 0; j < numberOfClasses; j++)
        {
            fprintf (stderr, "Q[%d][%d] = %f\n", i, j, gsl_matrix_get (Q, i, j));
        }
    }
    }
#endif

    ListDestroy (&list);
    list = NULL;

    *numClasses = numberOfClasses;

    return Q;
}


/*============================================================================
 *
 * Name:: ComputeXStar
 *
 * Description::
 *    Calculates the XStar matrix.
 *    (XStar = X*Q with dim N x |C| p.30 of T. Hastie paper)
 *
 * Inputs::
 *     X          - Gene expression matrix.
 *     n          - Number of rows (genes) in gene expression matrix.
 *     p          - Number of columns (samples/cells) in gene expr matrix.
 *     Q          - Classification matrix.
 *     numClasses - |C|
 *
 * Outputs::
 *     XStar      - Rows are standardized versions of class means for each gene.
 *
 * Return Value::
 *     None
 *
 *============================================================================*/

static void ComputeXStar (
  matrix_t *XStar,
  matrix_t *X,
  int n,
  int p,
  matrix_t *Q,
  int numClasses)
{
    register int i;

    assert (XStar != NULL);
    assert (X != NULL);
    assert (Q != NULL);

    /* XStar = X*Q */
    gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, X, Q, 0.0, XStar);
#ifdef NR
    for (i = 0; i < n; i++)
    {
        register int j;

        for (j = 0; j < numClasses; j++)
        {
            register int k;
            double temp = 0.0;

            for (k = 0; k < p; k++)
                temp += X[i][k] * Q[k][j];

            XStar[i][j] = temp;
        }
    }
#endif
}


/*============================================================================
 *
 * Name:: ComputeXtransposeX
 *
 * Description::
 *     Computes M=XT*X (X-transpose x X) Here: X[...][Jmax], M[Jmax][Jmax]
 *
 * Inputs::
 *     X        - X[...][Jmax] - Gene expression matrix.
 *     Jmax     - Number of columns (samples/cells).
 *     numGenes - Number of rows (genes).
 *     geneMap  - Vector[0..numGenes-1] of gene numbers.
 *
 * Outputs::
 *     M        - M[Jmax][Jmax]
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void ComputeXtransposeX (
    matrix_t *M,
    matrix_t *X,
    int Jmax,
    int numGenes,
    const ivector_t *geneMap)
{
    register int i;

    assert (X != NULL);
    assert (M != NULL);
    assert (numGenes > 0);
    assert (Jmax > 0);
    assert (geneMap != NULL);

    /* Find M=XT*X */
    for (i = 0; i < Jmax; i++)
    {
        register int j;

        for (j = 0; j <= i; j++)
        {
            register int k;
            double sum = 0.0;

            for (k = 0; k < numGenes; k++) {
		int gene = gsl_vector_int_get (geneMap, k);
                sum += gsl_matrix_get (X, gene, i) * gsl_matrix_get (X, gene, j);
	    }

            gsl_matrix_set (M, j, i, sum);
	    gsl_matrix_set (M, i, j, sum);
        }
    }
}


/*============================================================================
 *
 * Name:: ConstructLabels
 *
 * Description::
 *     Constructs an array of strings in the format: "xxx999zzz" where
 *     xxx is a prefix, 999 is current count, and zzz is a suffix.
 *
 * Inputs::
 *     numLabels - Number of labels to create.
 *     prefix    - String to prepend to label name.
 *     suffix    - String to append to label name.
 *
 * Outputs::
 *     labels    - Array where strings are stored.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void ConstructLabels (
  char **labels,
  int numLabels,
  const char *prefix,
  const char *suffix)
{
    register int i;
    char buffer[CVECTOR_ELEM_LEN+1];

    assert (labels != NULL);
    assert (numLabels > 0 && numLabels <= SHRT_MAX);
    assert (prefix != NULL);
    assert (suffix != NULL);

    for (i = 0; i < numLabels; i++)
    {
        snprintf (buffer, sizeof(buffer), "%s%05d%s", prefix, i+1, suffix);
        strcpy (labels[i], buffer);
    }
}


/*============================================================================
 *
 * Name:: ConstructQColumn
 *
 * Description::
 *     Constructs a single column in the Q matrix (Hastie, pg. 30) used in
 *     supervised shaving.
 *
 * Inputs::
 *     classNum - Column number of Q matrix.
 *     count    - Number of times 'classNum' appears within the
 *                classification file.
 *     indices  - Location indices of 'classNum' within the
 *                classification file (what line numbers within the file).
 *
 * Outputs::
 *     Q        - Matrix with modified 'classNum' column.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void ConstructQColumn (
  matrix_t *Q,
  unsigned classNum,
  unsigned count,
  const unsigned *indices)
{
    register int i;

    assert (Q != NULL);
    assert (indices != NULL);

    for (i = 0; i < count; i++)
    {
        int idx;

        idx = indices[i];
	fprintf (stderr, "ConstructQColumn: %d -> idx=%d class=%d\n", i, idx, classNum);
        gsl_matrix_set (Q, idx, classNum, 1.0 / sqrt ((double)count));
    }
}


/*============================================================================
 *
 * Name:: DisplayProgress
 *
 * Description::
 *     Writes "encoded" message to output where Java front-end is monitoring.
 *     Message contains property key for resource file and any arguments
 *     needed to translate the message to an appropriate format for display.
 *
 * Inputs::
 *     messageFormat - Format string (printf-like) of message to be displayed.
 *     args          - Variable argument list.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

/*VARARGS1*/
static void DisplayProgress (const char *messageFormat, ...)
{
    Message_t message;
    va_list args;

    assert (messageFormat != NULL);

    if (gShowProgress == false)
        return;

    PROCESS_VAR_ARGS(message,messageFormat,args);
    fprintf (stdout, "%s\n", message);
    fflush (stdout);
}


/*============================================================================
 *
 * Name::  DoNestedShave
 *
 * Description::
 *     Creates a nested sequence of clusters.
 *
 * Inputs::
 *     X          - Gene expression matrix.
 *     Q          - Supervision matrix, if any.
 *     numSamples - Number of columns in the gene expression matrix X.
 *     numClasses - Number of columns in the supervision matrix Q.
 *
 * Outputs::
 *     princomps   - Array of pointers to hold principal components or NULL.
 *     weights     - Matrix[0..numShavings-1][0..n-1] to store weights of
 *                   princomps.
 *     eigenvalues - Array to hold eigenvalues or NULL.
 *
 * Inputs / Outputs::
 *     result
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

/*
 * If any diagonal element of M is reduced to less than RECOMPUTE_M_PRECISION
 * times its original value, we may start to lose precision, so recompute M in
 * its entirety from X. It's probably OK for this value to be as large as half
 * the machine's precision, so the value defined above is very conservative.
 * Nevertheless, it nearly always eliminates the need to recompute M when X is
 * still large.
 */

static void DoNestedShave (
  Result_t *result,
  matrix_t *X,
  matrix_t *Q,
  int numSamples,
  int numClasses,
  vector_t **princomps,
  matrix_t *weights,
  vector_t *eigenvalues)
{
    register int i;
    int shaveNum;        /* shaveNum - loop index */
    matrix_t *M;
    vector_t *PC;
    vector_t *alpha;
    int n;
    int p;
    bool recomputeM;
    vector_t *checkDiag;
    matrix_t *XStar;            /* For supervised shaving only */

#ifdef TRACEFNS
    fprintf (stderr, "> DoNestedShave\n");
#endif
    assert (result != NULL);
    assert (X != NULL);
    assert (numSamples > 0);
    assert (weights != NULL);

    n = gsl_vector_int_get (result->clusterSize, 0);
    assert (n > 0);

    if (Q != NULL)
    {
        XStar = dmatrix (0, n-1, 0, numClasses-1);
        ComputeXStar (XStar, X, n, numSamples, Q, numClasses);
        CenterData (XStar, n, numClasses);
        p = numClasses;
    }
    else
    {
        XStar = X;        /* XStar is matrix to use in computations below */
        p = numSamples;
    }

    /* Initialization of arrays */
    alpha = dvector (0, n-1);
    M = dmatrix (0, p-1, 0, p-1);
    checkDiag = dvector (0, p-1);
    if (princomps == NULL)
        PC = dvector (0, numSamples-1);

    /* Initialize geneMap and cluster */
    for (i = 0; i < n; i++)
        gsl_vector_int_set (result->geneMap, i, i);

    /* Begin iterative shaving procedure */
    recomputeM = true;
    for (shaveNum = 0; shaveNum < result->numShavings; shaveNum++)
    {
        /* r : # Genes in current cluster */
        int r = gsl_vector_int_get (result->clusterSize, shaveNum);
        /* numLeft : # Genes in next cluster */
        int numLeft = gsl_vector_int_get (result->clusterSize, shaveNum+1);
        double eigenvalue;
	gsl_vector_view weight;

#if 0 && defined(DEBUG)
        fprintf (stderr, "  Shaving %d, genes remaining = %d\n", shaveNum, r);
#endif

        if (recomputeM)
        {
            ComputeXtransposeX (M, XStar, p, r, result->geneMap);
            for (i = 0; i < p; i++)
                gsl_vector_set (checkDiag, i, gsl_matrix_get (M, i, i) * RECOMPUTE_M_PRECISION);
            recomputeM = false;
        }

        if (princomps != NULL)
            princomps[shaveNum] = PC = dvector (0, numSamples-1);

	weight = gsl_matrix_row (weights, shaveNum);

        /*
         * For displaying signed mean genes in the display (Splus) code, we
         * need a PC vector of length numSamples, whether we're supervised or
         * not.
         */
        if (Q == NULL)
        {
            /* Unsupervised, we find the sample PCs directly */
            FindPC (M, p, PC, &eigenvalue);
            DotProd (result->geneMap, X, r, numSamples,
                     PC, &weight.vector, alpha);
        }
        else
        {
            /*
             * Supervised, we find the class PC, and use them for computing
             * alpha, then multiply by Q to get the sample PCs.
             */
            vector_t *ps =  dvector (0, numClasses-1);
            FindPC (M, numClasses, ps, &eigenvalue);
            DotProd (result->geneMap, XStar, r, numClasses,
                     ps, &weight.vector, alpha);

            if (princomps != NULL)
            {
                for (i = 0; i < numSamples; i++)
                {
                    register int j;
                    double total = 0.0;

                    for (j = 0; j < numClasses; j++)
                        total += gsl_matrix_get (Q, i, j) * gsl_vector_get (ps, j);

                    gsl_vector_set (PC, i, total);
                }
            }
            free_dvector (ps, 0, numClasses-1);
        }

        if (eigenvalues != NULL)
            gsl_vector_set (eigenvalues, shaveNum, eigenvalue);

        ShaveOff (alpha, result->geneMap, r, numLeft);
	{
	gsl_vector_int_view shaved = gsl_vector_int_subvector (result->geneMap, numLeft, n-numLeft);
        UpdateXtransposeX (M, XStar, p, r-numLeft, &shaved.vector);
	}
        for (i = 0; i < p; i++)
        {
            if (gsl_matrix_get (M, i, i) < gsl_vector_get (checkDiag, i))
            {
                recomputeM = true;
                break;
            }
        }
    }

    free_dvector (alpha, 0, n-1);
    free_dvector (checkDiag, 0, p-1);
    free_dmatrix (M, 0, p-1, 0, p-1);
    if (princomps == NULL)
        free_dvector (PC, 0, numSamples-1);
    if (Q != NULL)
        free_dmatrix (XStar, 0, n-1, 0, numClasses-1);
#ifdef TRACEFNS
    fprintf (stderr, "< DoNestedShave\n");
#endif
}


/*============================================================================
 *
 * Name:: DotProd
 *
 * Description::
 *     Computes the Inner product of each row of X with the Principal Component.
 *
 * Note::
 *    In this function n is not the number of genes - should rename it.
 *    X is a reduced matrix containing rows for only those genes in Index.
 *
 * Inputs::
 *     Index - Maps row numbers to gene numbers.
 *     X     - Gene expression matrix.
 *     n     - Number of raws (genes) in X and Index.
 *     p     - Number of columns (samples/cells) in X.
 *     PC    - Principal component vector [0..p-1] of X.
 *
 * Outputs::
 *     A      - Vector [0..n-1] with elements the dot products of the principal
 *              component with each row of the gene expression matrix.
 *     weight - Vector [0..numGenes-1] in which to store the value of each dot
 *              product.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void DotProd (
  ivector_t *Index,
  matrix_t *X,
  int n,
  int p,
  vector_t *PC,
  vector_t *weight,
  vector_t *A)
{
    int i;

#ifdef TRACEFNS
    fprintf (stderr, "> DotProd\n");
#endif
    assert (X != NULL);
    assert (n > 0);
    assert (p > 0);
    assert (PC != NULL);
    assert (A != NULL);

    for (i = 0; i < n; i++)
    {
        int j;
        double temp;
        int gene = gsl_vector_int_get (Index, i);

        temp = 0.0;
        for (j = 0; j < p; j++)
            temp += gsl_matrix_get (X, gene, j) * gsl_vector_get (PC, j);

        if (weight != NULL)
            gsl_vector_set (weight, gene, temp);

        gsl_vector_set (A, i, fabs (temp));
    }
#ifdef TRACEFNS
    fprintf (stderr, "< DotProd\n");
#endif
}


/*============================================================================
 *
 * Name:: EstimateOptimal
 *
 * Description::
 *     Performs a gene shaving of the genes in X, possibly supervised by Q.
 *     Finds the index of the Largest Gap vector element, performing the
 *     gap statistics, and stores the gap info into various files.
 *
 * Inputs::
 *     X           - Gene expression matrix.
 *     n           - Number of raws (genes) in X.
 *     p           - Number of columns (samples/cells) in X.
 *     nperm       - Number of randomizations for the gap statistics.
 *     tolerance   - Gap tolerance.
 *     count       - Current cluster number (in the extraction
 *                   process, i.e. loop index from GeneShave).
 *     numClusters - Total number of clusters to be extracted.
 *     Q           - Optional supervision matrix.
 *     numClasses  - Number of columns (classes) in Q.
 *
 * Outputs::
 *     (None)
 *
 * Inputs/Outputs::
 *     result   - On input, contains the number of shavings and the number
 *                genes in each shaving. On output, the other fields are
 *                completed.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void EstimateOptimal (
  Result_t *result,
  matrix_t *X,
  int n,
  int p,
  int nperm,
  double tolerance,
  int count,
  int numClusters,
  matrix_t *Q,
  int numClasses)
{
    register int i;
    int numShavings;
    vector_t *DK;
    vector_t *DKbar;
    vector_t *DKtemp;
    vector_t *VB;
    vector_t *VT;
    vector_t *VW;
    matrix_t *Db;
    /* For each shaving [ 0..numShavings-1 ], we keep pointers to */
    vector_t *eigenvalues;           /* the eigenvalue */
    vector_t **princomps;            /* the principal components */
    matrix_t *weights;              /* the gene weights */

    assert (result != NULL);
    assert (X != NULL);
    assert (n > 0);
    assert (p > 0);
    assert (nperm > 0);

    numShavings = result->numShavings;    /* Shorthand */
    assert (numShavings > 0);

    /* Allocate storage and initialize temporary data structures */
    eigenvalues = dvector (0, numShavings-1);
    princomps   = (vector_t **)pvector (result->numShavings);
    weights     = dmatrix (0, numShavings-1, 0, n-1);

    /* Compute initial nested shaving */
    DisplayProgress (SHAVE_INIT_NESTED_SEQUENCE);

    DoNestedShave (result, X, Q, p, numClasses,
                   princomps, weights, eigenvalues);

    DisplayProgress ("%s %d %d", ESTIMATE_OPT_CLUSTER_SIZE, count, numClusters);

    DK = dvector (0, numShavings-1);
    VB = dvector (0, numShavings-1);
    VT = dvector (0, numShavings-1);
    VW = dvector (0, numShavings-1);

    FindR2 (result, DK, X, n, p, weights, VB, VT, VW);

    Db = dmatrix (0, nperm-1, 0, numShavings-1);

    /* Start the randomized sequence */
    {
        matrix_t *tempWeights = dmatrix (0, numShavings-1, 0, n-1);
        vector_t *tempVB = dvector (0, numShavings-1);
        vector_t *tempVT = dvector (0, numShavings-1);
        vector_t *tempVW = dvector (0, numShavings-1);
        matrix_t *XX = dmatrix (0, n-1, 0, p-1);
        Result_t *tempResult = CloneResult (n, p, result);

        for (i = 0; i < nperm; i++)
        {
	    gsl_vector_view dbrow;

            DisplayProgress (RANDOMIZE_MATRIX);
            PermuteMatrixRowElements (X, XX, n, p);
            DisplayProgress ("%s %d %d",
                             SHAVE_RAND_NESTED_SEQUENCE, i+1, nperm);
            DoNestedShave (tempResult, XX, Q, p, numClasses,
                           NULL, tempWeights, NULL);
	    dbrow = gsl_matrix_row (Db, i);
            FindR2 (tempResult, &dbrow.vector, XX, n, p,
                    tempWeights, tempVB, tempVT, tempVW);

#if 0
            {
                register int j;

                fprintf (stderr, "\n************ i = %d\n", i);
                for (j = 0; j < numShavings; j++)
                    fprintf (stderr, "Db[%2d][%2d] = %f\n", i, j, gsl_matrix_get (Db, i, j));
            }
#endif
        }
        FreeResult (tempResult, n, p);
        tempResult = NULL;
        free_dmatrix (tempWeights, 0, numShavings-1, 0, n-1);
        free_dvector (tempVB, 0, numShavings-1);
        free_dvector (tempVT, 0, numShavings-1);
        free_dvector (tempVW, 0, numShavings-1);
        free_dmatrix (XX, 0, n-1, 0, p-1);
    }


    /* Find D_k_bar */
    DKbar = dvector (0, numShavings-1);
    for (i = 0; i < numShavings; i++)
    {
        register int j;
        double temp = 0.0;

        for (j = 0; j < nperm; j++)
            temp += gsl_matrix_get (Db, j, i);
        gsl_vector_set (DKbar, i, temp / nperm);
    }

    /* Find gap vector and index of maximal value */
    DKtemp = dvector (0, numShavings-1);
    {
        double max;

        max = gsl_vector_get (DK, 0) - gsl_vector_get (DKbar, 0);
	gsl_vector_set (DKtemp, 0, max);
        result->optIndex = 0;
        for (i = 1; i < numShavings; i++)
        {
            gsl_vector_set (DKtemp, i, gsl_vector_get (DK, i) - gsl_vector_get (DKbar, i));
            if (gsl_vector_get (DKtemp, i) >= max)
            {
                max = gsl_vector_get (DKtemp, i);
                result->optIndex = i;
            }
#if 0 && defined(DEBUG)
            fprintf (stderr, "  DKbar[%d] = %f\t DK[%d] = %f\t gap[%d] = %f\n",
                     i, gsl_vector_get (DKbar,i),
                     i, gsl_vector_get (DK, i),
                     i, gsl_vector_get (DKtemp, i));
#endif
        }
    }

    /* Apply gap tolerance */
    {
        double threshold = (1.0 - tolerance/100.0) * gsl_vector_get (DKtemp, result->optIndex);
        for (i = result->optIndex-1; i >= 0; i--)
        {
            if (gsl_vector_get (DKtemp, i) > threshold)
                result->optIndex = i;
        }
    }

    {
    int j;

    StoreVarianceInfo (count, numShavings, nperm, result->clusterSize,
                       DK, DKbar, VB, VT, VW, Db);
    for (i = 0; i < nperm; i++)
	for (j = 0; j < numShavings; j++)
	    gsl_matrix_set (Db, i, j, gsl_vector_get (DK, j) - gsl_matrix_get (Db, i, j));
    StoreGapInfo (count, numShavings, nperm, result->clusterSize, DKtemp, Db);
    }

    /* Store results */
    result->eigenvalue = gsl_vector_get (eigenvalues, result->optIndex);
    {
        vector_t *pc = princomps[result->optIndex];

        for (i = 0; i < p; i++)
            gsl_vector_set (result->principalComponent, i, gsl_vector_get (pc, i));

        for (i = 0; i < n; i++)
            gsl_vector_set (result->weight, i, gsl_matrix_get (weights, result->optIndex, i));
    }

    /* Free memory */
    free_dvector (DK, 0, numShavings-1);
    free_dvector (DKbar, 0, numShavings-1);
    free_dvector (DKtemp, 0, numShavings-1);
    free_dvector (VB, 0, numShavings-1);
    free_dvector (VT, 0, numShavings-1);
    free_dvector (VW, 0, numShavings-1);
    free_dmatrix (Db, 0, nperm-1, 0, numShavings-1);
    {
        for (i = 0; i < numShavings; i++)
            free_dvector (princomps[i], 0, p-1);
    }
    free_pvector ((void **)princomps, numShavings);
    free_dmatrix (weights, 0, numShavings-1, 0, n-1);
    free_dvector (eigenvalues, 0, n-1);
}


/*============================================================================
 *
 * Name:: FindCluster
 *
 * Description::
 *     Creates a nested sequence of clusters, finds the optimal cluster in the
 *     sequence using gap statistics, and stores the cluster information of the
 *     original genes to a text file. Computes the average signed gene using
 *     the current matrix values, and stores it in an output file.
 *     Orthogonalizes the current matrix with respect to the average signed
 *     gene.
 *
 * Inputs::
 *     originalX     - Original gene expression matrix.
 *     n             - Number of raws (genes) in X.
 *     p             - Number of columns (samples/cells) in X.
 *     shavePercent  - Percentile to shave.
 *     nperm         - Number of randomizations for the gap statistics.
 *     tolerance     - Gap tolerance.
 *     count         - Cluster number (in the main loop in GeneShave).
 *     numClusters   - Total number of clusters to extract.
 *     GeneLbls      - String arrays of gene labels.
 *     point         - String arrays of cluster filenames.
 *     clbl          - String arrays of cluster label filenames.
 *     Q             - Classification matrix, if supervised.
 *     numClasses    - Number of classifications, if supervised.
 *
 * Outputs::
 *     (None)
 *
 * Input/Outputs::
 *     result        - Pointer to Result_t containing optimal index, eigenvalue,
 *                     principal component, and number of genes in extracted
 *                     cluster.
 *     X             - Gene expression matrix.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void FindCluster (
  matrix_t *X,
  matrix_t *originalX,
  int n,
  int p,
  double shavePercent,
  int nperm,
  double tolerance,
  int count,
  int numClusters,
  char **GeneLbls,
  char **point,
  char **clbl,
  matrix_t *Q,
  int numClasses,
  Result_t *result)
{
    vector_t *x_bar_opt;

    assert (X != NULL);
    assert (originalX != NULL);
    assert (n > 1);
    assert (p > 1);
    assert (shavePercent > 0.0 && shavePercent < 1.0);
    assert (nperm > 0);
    assert (count >= 1);
    assert (numClusters >= 1);
    assert (GeneLbls != NULL);
    assert (point != NULL);
    assert (clbl != NULL);
    assert (result != NULL);

    EstimateOptimal (result, X, n, p, nperm, tolerance,
                     count, numClusters, Q, numClasses);

    StoreClusterInfo (originalX, p, gsl_vector_int_get (result->clusterSize, result->optIndex),
                      result->geneMap, clbl, GeneLbls, point,
                      result->weight, count);

    /* Initialize and find x_bar_opt */
    x_bar_opt = dvector (0, p-1);
    CalcAvgSignedGene (X, p, gsl_vector_int_get (result->clusterSize, result->optIndex),
                       x_bar_opt, result->geneMap, result->weight);

    DisplayProgress ("%s %d", ORTHOGONALIZE_MATRIX, count);
    Orthogonalize (X, x_bar_opt, n, p);

    /* Free memory */
    free_dvector (x_bar_opt, 0, p-1);
}


/*============================================================================
 *
 * Name::  FindPC
 *
 * Description::
 *     Finds the principal component using singular value decomposition.
 *
 * Inputs::
 *     M - Square matrix.
 *     p - Number of rows/columns in M.
 *
 * Outputs::
 *     PC         - Principal component vector.
 *     eigenvalue - Pointer to where eigenvalue will be stored.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void FindPC (
  matrix_t *M,
  int p,
  vector_t *PC,
  double *eigenvalue)
{
    register int i;
    int maxind;
    double max = 0.0;
    matrix_t *U;
    matrix_t *V;
    vector_t *D;

#ifdef TRACEFNS
    fprintf (stderr, "> FindPC\n");
#endif
    assert (M != NULL);
    assert (p > 0);
    assert (PC != NULL);
    assert (eigenvalue != NULL);

    D = dvector (0, p-1);
    U = dmatrix (0, p-1, 0, p-1);
    V = dmatrix (0, p-1, 0, p-1);


    /* Find svd(M) */
    svdcmp (M, U, p, p, D, V);

    maxind = 1;
    max = gsl_vector_get (D, 0);
    fprintf (stderr, "  Eigenvalue = %g\n", max);
    for (i = 2; i <= p; i++)
    {
        if ((gsl_vector_get (D, i-1)-max) > DBL_EPSILON)
        {
            max = gsl_vector_get (D, i-1);
            maxind = i;
        }
    }

    if (gsl_vector_get (D, maxind-1) < 1.0e40) {
        *eigenvalue = gsl_vector_get (D, maxind-1);
        for (i = 1; i <= p; i++)
            gsl_vector_set (PC, i-1, -gsl_matrix_get (U, i-1, maxind-1)); /* - needed! */
    }
    else {
        *eigenvalue = 0.0;
        for (i = 1; i <= p; i++)
            gsl_vector_set (PC, i-1, 0.0);
    }

#if 0
    fprintf (stderr, "  Eigenvalue = %g\n", max);
    fprintf (stderr, "  PC =");
    for (i = 0; i < p; i++)
	fprintf (stderr, " %5g", gsl_vector_get (PC, i));
    fprintf (stderr, "\n");
#endif
    free_dvector (D, 0, p-1);
    free_dmatrix (U, 0, p-1, 0, p-1);
    free_dmatrix (V, 0, p-1, 0, p-1);
#ifdef TRACEFNS
    fprintf (stderr, "< FindPC\n");
#endif
}


/*============================================================================
 *
 * Name:: FindR2
 *
 * Description::
 *     Calculates the R2 measure for a given nested sequence of clusters.
 *
 * Inputs::
 *     result     - Nested sequence of clusters to analyze.
 *     X          - Gene expression matrix.
 *     n          - Number of rows (genes) in X.
 *     p          - Number of columns (samples/cells) in X.
 *     weights    - Matrix [0..result->numShavings-1][0..n-1] giving the weight
 *                  of the PC of each gene in each cluster
 *
 * Outputs::
 *     varBetween - Between variance.
 *     varTotal   - Total variance.
 *     varWithin  - Within variance.
 *
 * Inputs / Outputs::
 *     R2         - Vector with entries containing the R2 measure for each
 *                  cluster in the nested sequence.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void FindR2 (
  Result_t *result,
  vector_t *R2,
  matrix_t *X,
  int n,
  int p,
  matrix_t *weights,
  vector_t *varBetween,
  vector_t *varTotal,
  vector_t *varWithin)
{
    register int i;
    vector_t *X_j_bar;

#ifdef TRACEFNS
    fprintf (stderr, "> FindR2\n");
#endif
    assert (result != NULL);
    assert (R2 != NULL);
    assert (X != NULL);
    assert (n > 0);
    assert (p > 0);
    assert (varBetween != NULL);
    assert (varTotal != NULL);
    assert (varWithin != NULL);

    X_j_bar = dvector (0, p-1);

    /* For each cluster... */
    for (i = 0; i < result->numShavings; i++)
    {
        register int j;
        register int k;
        double x_bar = 0.0;
        double VB = 0.0;
        double VT = 0.0;
        double VW = 0.0;
	gsl_vector_view weight = gsl_matrix_row (weights, i);

        k = gsl_vector_int_get (result->clusterSize, i);

        CalcAvgSignedGene (X, p, k, X_j_bar, result->geneMap, &weight.vector);

        /* Find x_bar */
        x_bar = 0.0;
        for (j = 0; j < p; j++)
            x_bar += gsl_vector_get (X_j_bar, j);
        x_bar /= p;

        /* Find Between Variance */
        {
            VB = 0.0;
            for (j = 0; j < p; j++)
            {
                VB += gsl_pow_2 (gsl_vector_get (X_j_bar, j));
            }
            VB /= p;
            VB -= x_bar*x_bar;
        }

        /* Find Total Variance */
        {
            register int l;
            register int m;

            VT = 0.0;
            for (l = 0; l < k; l++)
            {
                m = gsl_vector_int_get (result->geneMap, l);
                for (j = 0; j < p; j++)
                {
                    VT += gsl_pow_2 (gsl_matrix_get (X, m, j));
                }
            }
            VT /= k * p;
            VT -= x_bar*x_bar;
        }

        /* Find Within Variance */
        VW = VT - VB;

        /* Construct R2 vector */
        if (VT < 5*DBL_MIN)
            gsl_vector_set (R2, i, 0.00);
        else
            gsl_vector_set (R2, i, 100.0 * VB / VT);

        /* Save results */
        gsl_vector_set (varBetween, i, VB);
        gsl_vector_set (varTotal, i, VT);
        gsl_vector_set (varWithin, i, VW);

#if 0 && defined(DEBUG)
        fprintf (stderr, "R2[%d] = %f   (100.0 * VB[%f] / VT[%f])\n",
                 i, R2[i], VB, VT);
#endif
    }

    free_dvector (X_j_bar, 0, p-1);
#ifdef TRACEFNS
    fprintf (stderr, "< FindR2\n");
#endif
}


/*============================================================================
 *
 * Name:: FreeResult
 *
 * Description::
 *     Deallocates a result structure, including all its component vectors.
 *
 * Inputs::
 *     result - Result structure to deallocate.
 *     N      - Number of genes.
 *     p      - Number of samples per gene.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void FreeResult (Result_t *result, int N, int p)
{
    assert (result != NULL);
    assert (N > 0);
    assert (p > 0);

    free_dvector (result->principalComponent, 0, p-1);
    free_dvector (result->weight, 0, N-1);
    free_ivector (result->geneMap, 0, N-1);
    free_ivector (result->clusterSize, 0, result->numShavings);

#ifdef DEBUG
    memset (result, '\0', sizeof (Result_t));
#endif

    free (result);
}


/*============================================================================
 *
 * Name:: IsDirectory
 *
 * Description::
 *     Determines whether the argument is a directory.
 *
 * Inputs::
 *     pathName - Path name to have its file-type bits examined.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     Returns true if pathname is a directory; otherwise, false.
 *
 *============================================================================*/

static bool IsDirectory (const char *pathName)
{
    struct stat buf;

    assert(pathName != NULL);

    if (!((stat (pathName, &buf) == -1) ||
          (S_ISDIR (buf.st_mode) == 0)))
        return true;

    return false;
}


/*============================================================================
 *
 * Name:: IsRegularFile
 *
 * Description::
 *     Determines whether the argument is a regular file.
 *
 * Inputs::
 *     pathName - Path name to have its file-type bits examined.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     Returns true if pathname is a regular file; otherwise, false.
 *
 *============================================================================*/

static bool IsRegularFile (const char *pathName)
{
    struct stat buf;

    assert(pathName != NULL);

    if (!((stat (pathName, &buf) == -1) ||
          (S_ISREG (buf.st_mode) == 0)))
        return true;

    return false;
}


/*============================================================================
 *
 * Name:: NewResult
 *
 * Description::
 *     Allocates a new result structure, computes the number and sizes of
 *     the shavings in the cluster, and allocates all component vectors of
 *     the structure.
 *
 * Inputs::
 *     N          - Number of genes.
 *     p          - Number of samples per gene.
 *     percentage - Percentage of genes to shave at each step.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     Returns pointer to newly created instance of Result_t.
 *
 *============================================================================*/

static Result_t *NewResult (int N, int p, double percentage)
{
    Result_t *result;
    register int i;

    assert (N > 0);
    assert (p > 0);

    result = (Result_t *) calloc (1, sizeof(Result_t));
    if (result == NULL)
    {
         ErrMessage_t errMessage;

         sprintf (errMessage, "Error creating result structure: %s\n",
                  strerror (errno));
         Abend (errMessage);
         /*NOTREACHED*/
    }

    /* Compute number of shavings */
    result->numShavings = 0;
    for (i = N; i > 1; i -= ShaveCount (i, percentage))
        result->numShavings++;

    /* Compute size of each shaving */
    {
        register int r;

        result->clusterSize = ivector (0, result->numShavings);
        r = N;
	gsl_vector_int_set (result->clusterSize, 0, r);
        for (i = 0; i < result->numShavings; ) {
            r -= ShaveCount (r, percentage);
	    gsl_vector_int_set (result->clusterSize, ++i, r);
        }
    }

    result->geneMap = ivector (0, N-1);
    result->weight  = dvector (0, N-1);
    result->principalComponent = dvector (0, p-1);

    return result;
}


/*============================================================================
 *
 * Name:: Orthogonalize
 *
 * Description::
 *     Orthogonalizes the gene expression matrix (in the sense of regression)
 *     w.r.t. the mean gene of the already found cluster.
 *
 * Inputs::
 *     x_bar_opt - Mean gene of the already found cluster.
 *     n         - Number of rows (genes) in gene expression matrix.
 *     p         - Number of columns (samples/cells) in gene expression matrix.
 *
 * Outputs::
 *     (None)
 *
 * Inputs / Outputs::
 *     X - Gene expression matrix.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void Orthogonalize (
  matrix_t *X,
  vector_t *x_bar_opt,
  int n,
  int p)
{
    matrix_t *yty;
    matrix_t *new;
    double scale;
    register int i;

    assert (X != NULL);
    assert (x_bar_opt != NULL);
    assert (n > 0);
    assert (p > 0);

    yty = dmatrix (0, p-1, 0, p-1);
    new = dmatrix (0, n-1, 0, p-1);

    /* Determine scale */
    {
        double mag = 0.0;

        for (i = 0; i < p; i++)
            mag += gsl_pow_2 (gsl_vector_get (x_bar_opt, i));

#ifdef DEBUG
        fprintf (stderr, "  Orthogonalize: Mag = %g\n", mag);
#endif
        scale = -1.0 / mag;
    }

    for (i = 0; i < p; i++)
    {
        register int j;
        double si = scale * gsl_vector_get (x_bar_opt, i);

        for (j = 0; j < i; j++) {
	    double v = si * gsl_vector_get (x_bar_opt, j);
            gsl_matrix_set (yty, j, i, v);
	    gsl_matrix_set (yty, i, j, v);
	}
        gsl_matrix_set (yty, i, i, 1.0 + si * gsl_vector_get (x_bar_opt, i));
    }

    gsl_blas_dgemm (CblasNoTrans, CblasNoTrans, 1.0, X, yty, 0.0, new);
#ifdef NR
    for (i = 0; i < n; i++)
    {
        register int j;

        for (j = 0; j < p; j++)
        {
            register int k;
            double temp = 0.0;

            for (k = 0; k < p; k++)
                temp += X[i][k] * yty[k][j];
            new[i][j] = temp;
        }
    }
#endif

    gsl_matrix_memcpy (X, new);
#ifdef NR
    for (i = 0; i < n; i++)
    {
        register int j;

        for (j = 0; j < p; j++)
            X[i][j] = new[i][j];
    }
#endif
    free_dmatrix (yty, 0, p-1, 0, p-1);
    free_dmatrix (new, 0, n-1, 0, p-1);
}


/*============================================================================
 *
 * Name:: PermuteMatrixRowElements
 *
 * Description::
 *     Creates a matrix from the original with row elements randomly permuted.
 *
 * Inputs::
 *     M - Matrix used as source for row elements permutation.
 *     n - Number of rows.
 *     p - Number of columns.
 *
 * Outputs::
 *     permuted - Copy of M with permuted row elements.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void PermuteMatrixRowElements (
  matrix_t *M,
  matrix_t *permuted,
  int n,
  int p)
{
    register int i;
    Perm_t *perm;

    assert (M != NULL);
    assert (permuted != NULL);
    assert (n > 0);
    assert (p > 0);

    if ((perm = (Perm_t *)calloc (p, sizeof(Perm_t))) == NULL)
    {
        Abend ("allocation failure in calloc()");
        /*NOTREACHED*/
    }

    for (i = 0; i < n; i++)
    {
        register int j;

        for (j = 0; j < p; j++)
        {
            perm[j].index = j;
            perm[j].val = rand ();
        }

        qsort (perm, p, sizeof(Perm_t), ComparePerm);

        for (j = 0; j < p; j++)
            gsl_matrix_set (permuted, i, j, gsl_matrix_get (M, i, perm[j].index));
    }

    free (perm);
    perm = NULL;
}


/*============================================================================
 *
 * Name:: ShaveCount
 *
 * Description::
 *     Determines the number of genes to shave at a shaving step.
 *
 * Inputs::
 *     n - Current number of rows (genes) in the gene expression matrix.
 *     k - Proportion of genes to shave at each step.
 *
 * Outputs::
 *     (None)
 *
 * Inputs / Outputs::
 *     (None)
 *
 * Return Value::
 *     The number of genes to shave.
 *
 *============================================================================*/

static int ShaveCount (int n, double k)
{
    int ind_int;

    /*
     * Rate inhibit once number of genes shrink to certain size
     * in order not to shave off most interesting too rapidly.
     */
    if (n > SHAVING_THRESHOLD)
    {
        double ind;
        double fractionalPart;
        double integralPart;

        /* Round index at midpoint */
        ind = n * k;
        fractionalPart = modf (ind, &integralPart);
        if (fractionalPart > 0.5 || integralPart == 0)
            ind_int = (int) ceil (ind);
        else
            ind_int = (int) floor (ind);
        assert (ind_int >= 1);
    }
    else
        ind_int = 1;

    return ind_int;
}


/*============================================================================
 *
 * Name:: ShaveOff
 *
 * Description::
 *     Rearranges A such that the smallest n-r elements are at the end.
 *     Index is rearranged in the same way as A.
 *
 * Inputs::
 *     n - Number of rows (genes) in Index/A.
 *     r - Number of rows (genes) to keep.
 *
 * Outputs::
 *     (None)
 *
 * Inputs / Outputs::
 *     A     - Vector [0..n-1] containing the weight of each element in Index.
 *             (These are the dot products of the principal component
 *             with corresponding row of the gene expression matrix.)
 *     Index - Vector for storage of the "row numbers" of the genes remaining
 *             after the current shave.
 *
 * Return Value::
 *     None.
 *
 *============================================================================*/

static void ShaveOff (
  vector_t *A,
  ivector_t *Index,
  int n,
  int r)
{
    register int nmi;

#ifdef TRACEFNS
    fprintf (stderr, "> ShaveOff\n");
#endif
    assert (A != NULL);
    assert (Index != NULL);
    assert (n > 0);
    assert (r > 0);

    /* Loop over every element in the region being excluded */
    for (nmi = n-1; nmi >= r; nmi--)
    {
        int minIndex;
        double min;

        /* Find the smallest remaining element */
        {
            register int j;

            min = gsl_vector_get (A, nmi);
            minIndex = nmi;
            for (j = 0; j < nmi; j++)
            {
                if (gsl_vector_get (A, j) < min)
                {
                    min = gsl_vector_get (A, j);
                    minIndex = j;
                }
            }
        }

        /* Swap smallest remaining element with the element being excluded */
        if (minIndex != nmi)
        {
	    gsl_vector_swap_elements (A, minIndex, nmi);
	    gsl_vector_int_swap_elements (Index, minIndex, nmi);
        }
    }
#ifdef TRACEFNS
    fprintf (stderr, "< ShaveOff\n");
#endif
}


/*============================================================================
 *
 * Name:: StoreClusterInfo
 *
 * Description::
 *     Stores the gene labels and the cluster label filenames into files.
 *
 * Inputs::
 *     X           - Gene expression matrix.
 *     p           - Number of columns (samples/cells) in X.
 *     clusterSize - Cluster size.
 *     geneMap     - Vector storing the row numbers of genes in cluster.
 *     clbl        - String arrays of cluster label filenames.
 *     GeneLbls    - String arrays of gene labels.
 *     point       - String arrays of cluster filenames
 *     weight      - Vector[genes] holding weight of each gene in cluster.
 *     count       - Cluster number (in the main loop in GeneShave).
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void StoreClusterInfo (
  matrix_t *X,
  int p,
  int clusterSize,
  const ivector_t *geneMap,
  char **clbl,
  char **GeneLbls,
  char **point,
  const vector_t *weight,
  int count)
{
    assert (X != NULL);
    assert (p > 0);
    assert (clusterSize > 0);
    assert (geneMap != NULL);
    assert (clbl != NULL);
    assert (GeneLbls != NULL);
    assert (point != NULL);
    assert (weight != NULL);

    {
        register int i;
        PathName_t clusterExtraPathName;
	FILE *file;

        sprintf (clusterExtraPathName, "%s/%s-extra",
                 GetOutputDirectory (), point[count-1]);

	if ((file = fopen (clusterExtraPathName, "w")) == NULL)
	{
	    ErrMessage_t errMessage;

	    sprintf (errMessage, "StoreClusterInfo(\"%s\"): %s\n",
		     clusterExtraPathName,
		     strerror (errno));
	    Abend (errMessage);
	    /*NOTREACHED*/
	}
#ifdef DEBUG
	fprintf (stderr, "  Genes in cluster %d = %d:", count, clusterSize);
#endif
        for (i = 0; i < clusterSize; i++) {
	    int gene = gsl_vector_int_get (geneMap, i);
            fprintf (file, "%s %3d   %g\n",
                     GeneLbls[gene], gene+1, gsl_vector_get(weight,gene));
	}
        fclose (file);
    }

    /* Begin storing the cluster rows in a file */
    {
	    PathName_t clusterMatrixPathName;

        sprintf (clusterMatrixPathName, "%s/%s",
                 GetOutputDirectory (), point[count-1]);
        DataDump (X, p, clusterSize, geneMap, clusterMatrixPathName);
    }

    {
        PathName_t clusterLabelsPathName;

        sprintf (clusterLabelsPathName, "%s/%s",
                 GetOutputDirectory (), clbl[count-1]);
        StoreLabels (GeneLbls, clusterSize, geneMap, clusterLabelsPathName);
    }
}


/*============================================================================
 *
 * Name:: StoreEigenvalue
 *
 * Description::
 *     Creates eigenvalue file and writes some information in it.
 *     It then updates the eigen labels file with the name of each eigenvalue
 *     file that it creates. This particular routine follows the same format
 *     as the rest of the program in order to simplify S-plus processing later.
 *
 * Inputs::
 *     clusterNum - Number of the cluster being processed.
 *     eigenvalue - Eigenvalue for k-th cluster in sequence.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void StoreEigenvalue (
  int clusterNum,
  double eigenvalue)
{
    FILE *eigenvalueFile;
    PathName_t eigenvalueFileName;
    PathName_t eigenvaluePathName;

    sprintf (eigenvalueFileName, "%s%05d%s",
             EIGENVALUE_FILE_PREFIX, clusterNum, FILE_EXTENSION_SUFFIX);
    sprintf (eigenvaluePathName, "%s/%s",
             GetOutputDirectory (), eigenvalueFileName);

    if ((eigenvalueFile = fopen (eigenvaluePathName, "w")) != NULL)
    {
        fprintf (eigenvalueFile, "%f\n", eigenvalue);
        fclose (eigenvalueFile);

        /* Update its label file so information just written can be found */
        {
            static bool firstTime = true;
            static PathName_t labelsPathName;
            FILE *eigenvalueLabelsFile;
            char *mode = "a";

            if (firstTime)
            {
                mode = "w";
                sprintf (labelsPathName, "%s/%s",
                         GetOutputDirectory (), FILE_EIGENVALUE_LABELS);
                firstTime = false;
            }

            if ((eigenvalueLabelsFile = fopen (labelsPathName, mode)) != NULL)
            {
                fprintf (eigenvalueLabelsFile, "%s\n", eigenvalueFileName);
                fclose (eigenvalueLabelsFile);
            }
            else
                perror (labelsPathName);
        }
    }
    else
        perror (eigenvaluePathName);
}


/*============================================================================
 *
 * Name:: StoreGapInfo
 *
 * Description::
 *     Creates gap statistics file and writes some information in it.
 *     It then updates the gap labels file with the name of each gap
 *     statistics file that it creates.
 *
 * Inputs::
 *     clusterNum - Number of the cluster being processed.
 *     numShaves  - Number of shaves performed on that cluster.
 *     numGenesPerCluster - Vector containing number of genes per cluster.
 *     DKtemp     - Vector containing gap measure for k-th cluster in sequence.
 *     Db     - Matrix containing gap measure for k-th cluster in sequence.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void StoreGapInfo (
  int clusterNum,
  int numShaves,
  int numPerm,
  const ivector_t *numGenesPerCluster,
  const vector_t *DKtemp,
  const matrix_t *Db)
{
    register int i;
    FILE *gapFile;
    PathName_t gapFileName;
    PathName_t gapPathName;

    sprintf (gapFileName, "%s%05d%s",
             GAP_FILE_PREFIX, clusterNum, FILE_EXTENSION_SUFFIX);
    sprintf (gapPathName, "%s/%s",
             GetOutputDirectory (), gapFileName);

    if ((gapFile = fopen (gapPathName, "w")) != NULL)
    {
        for (i = 0; i < numShaves; i++)
        {
	    int j;

            fprintf (gapFile, "%d\t%f",
                     gsl_vector_int_get (numGenesPerCluster, i),
		     gsl_vector_get (DKtemp, i));
	    for (j = 0; j < numPerm; j++)
		fprintf (gapFile, "\t%f", gsl_matrix_get (Db, j, i));
	    fprintf (gapFile, "\n");
        }

        fclose (gapFile);

        /* Update its label file so information just written can be found */
        {
            static bool firstTime = true;
            static PathName_t labelsPathName;
            FILE *gapLabelsFile;
            char *mode = "a";

            if (firstTime)
            {
                mode = "w";
                sprintf (labelsPathName, "%s/%s",
                         GetOutputDirectory (), FILE_GAP_LABELS);
                firstTime = false;
            }

            if ((gapLabelsFile = fopen (labelsPathName, mode)) != NULL)
            {
                fprintf (gapLabelsFile, "%s\n", gapFileName);
                fclose (gapLabelsFile);
            }
            else
                perror (labelsPathName);
        }
    }
    else
        perror (gapPathName);
}


/*============================================================================
 *
 * Name:: StoreLabels
 *
 * Description::
 *     Stores an array of strings into a text file named 'filename'.
 *
 * Inputs::
 *     strings    - Array of strings to store.
 *     numStrings - Number of strings to save.
 *     map        - Optional index map giving array indices to store.
 *     filename   - String to prepend to label name.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void StoreLabels (
  char **strings,
  int numStrings,
  const ivector_t *map,
  const char *filename)
{
    FILE *file;
    register int i;

    assert (strings != NULL);
    assert (numStrings > 0);
    assert (filename != NULL);

    if ((file = fopen (filename, "w")) == NULL)
    {
        ErrMessage_t errMessage;

        sprintf (errMessage, "StoreLabels(\"%s\"): %s\n",
                 filename,
                 strerror (errno));
        Abend (errMessage);
        /*NOTREACHED*/
    }

    for (i = 0; i < numStrings; i++)
        fprintf (file, "%s\n", strings[(map == NULL) ? i : gsl_vector_int_get (map, i)]);

    fclose (file);
}


/*============================================================================
 *
 * Name:: StoreMatrix
 *
 * Description::
 *     Creates original matrix file and writes matrix in it.
 *
 * Inputs::
 *     X    - Gene expression matrix.
 *     rows - Number of genes (rows in gene expr matrix).
 *     cols - Number of samples (cell lines).
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void StoreMatrix (
  matrix_t *X,
  int rows,
  int cols)
{
    FILE *origMatrixFile;
    PathName_t origMatrixPathName;

    sprintf (origMatrixPathName, "%s/%s",
             GetOutputDirectory (), FILE_ORIG_MATRIX_VALUES);
    if ((origMatrixFile = fopen (origMatrixPathName, "w")) == NULL)
    {
        ErrMessage_t errMessage;

        sprintf (errMessage, "GeneShave(\"%s\"): %s\n",
                 origMatrixPathName,
                 strerror (errno));
        Abend (errMessage);
        /*NOTREACHED*/
    }
    else
    {
        register unsigned row;

        for (row = 0; row < rows; row++)
        {
            register unsigned col;

            for (col = 0; col < cols; col++)
            {
                fprintf (origMatrixFile, " %g", gsl_matrix_get (X, row, col));
            }
            fprintf (origMatrixFile, "\n");
        }

        fclose (origMatrixFile);
    }
}


/*============================================================================
 *
 * Name:: StorePrincipalComponent
 *
 * Description::
 *     Creates principal component file and writes some information in it.
 *     It then updates the principal component labels file with the name of
 *     each principal component file that it creates.
 *
 * Inputs::
 *     clusterNum - Number of the cluster being processed.
 *     pcSize     - Number of elements in the principal component vector.
 *     PC         - Vector containing principal component for k-th cluster in
 *                  sequence.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void StorePrincipalComponent (
  int clusterNum,
  size_t pcSize,
  const vector_t *PC)
{
    register int i;
    FILE *principalComponentFile;
    PathName_t principalComponentFileName;
    PathName_t principalComponentPathName;

    sprintf (principalComponentFileName, "%s%05d%s",
            PRINCIPAL_COMPONENT_FILE_PREFIX, clusterNum, FILE_EXTENSION_SUFFIX);
    sprintf (principalComponentPathName, "%s/%s",
             GetOutputDirectory (), principalComponentFileName);

    principalComponentFile = fopen (principalComponentPathName, "w");
    if (principalComponentFile != NULL)
    {
        for (i = 0; i < pcSize; i++)
        {
            fprintf (principalComponentFile, "%f\n", gsl_vector_get (PC, i));
        }

        fclose (principalComponentFile);

        /* Update its label file so information just written can be found */
        {
            static bool firstTime = true;
            static PathName_t labelsPathName;
            FILE *principalComponentLabelsFile;
            char *mode = "a";

            if (firstTime)
            {
                mode = "w";
                sprintf (labelsPathName, "%s/%s",
                         GetOutputDirectory (),FILE_PRINCIPAL_COMPONENT_LABELS);
                firstTime = false;
            }

            principalComponentLabelsFile = fopen (labelsPathName, mode);
            if (principalComponentLabelsFile != NULL)
            {
                fprintf (principalComponentLabelsFile, "%s\n",
                         principalComponentFileName);
                fclose (principalComponentLabelsFile);
            }
            else
                perror (labelsPathName);
        }
    }
    else
        perror (principalComponentPathName);
}


/*============================================================================
 *
 * Name:: StoreVarianceInfo
 *
 * Description::
 *     Creates variance statistics file and writes some information in it.
 *     It then updates the gap labels file with the name of each variance
 *     statistics file that it creates.
 *
 * Inputs::
 *     clusterNum - Number of the cluster being processed.
 *     numShaves  - Number of shaves performed on that cluster.
 *     numGenesPerCluster - Vector containing number of genes per cluster.
 *     DK         - Vector containing R2 measure for k-th cluster in sequence.
 *     DKbar      - Vector containing mean R2 measure for k-th cluster in
 *                  sequence taken over requested number of permutations.
 *     VB         - Vector containing between variance for k-th cluster in
 *                  sequence.
 *     VT         - Vector containing total variance for k-th cluster in
 *                  sequence.
 *     VW         - Vector containing within variance for k-th cluster in
 *                  sequence.
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void StoreVarianceInfo (
  int clusterNum,
  int numShaves,
  int nPerm,
  const ivector_t *numGenesPerCluster,
  const vector_t *DK,
  const vector_t *DKbar,
  const vector_t *VB,
  const vector_t *VT,
  const vector_t *VW,
  const matrix_t *Db)
{
    FILE *varianceFile;
    PathName_t varianceFileName;
    PathName_t variancePathName;

    assert (numGenesPerCluster != NULL);
    assert (DK != NULL);
    assert (DKbar != NULL);
    assert (VB != NULL);
    assert (VT != NULL);
    assert (VW != NULL);

    sprintf (varianceFileName, "%s%05d%s",
             VARIANCE_FILE_PREFIX, clusterNum, FILE_EXTENSION_SUFFIX);
    sprintf (variancePathName, "%s/%s",
             GetOutputDirectory (), varianceFileName);

    if ((varianceFile = fopen (variancePathName, "w")) != NULL)
    {
        /* Write variance file information to disk */
        {
            register int i;

            for (i = 0; i < numShaves; i++)
            {
	        int j;

                fprintf (varianceFile, "%d\t%f\t%f\t%f\t%f\t%f",
                         gsl_vector_int_get (numGenesPerCluster, i),
			 gsl_vector_get (DK, i),
			 gsl_vector_get (DKbar, i),
                         gsl_vector_get (VB, i),
			 gsl_vector_get (VT, i),
			 gsl_vector_get (VW, i));
	        for (j = 0; j < nPerm; j++)
		    fprintf (varianceFile, "\t%f", gsl_matrix_get (Db, j, i));
		fprintf (varianceFile, "\n");
            }
        }

        fclose (varianceFile);

        /* Update its label file so information just written can be found */
        {
            static bool firstTime = true;
            static PathName_t labelsPathName;
            FILE *labelsFile;
            char *mode = "a";

            if (firstTime)
            {
                mode = "w";
                sprintf (labelsPathName, "%s/%s",
                         GetOutputDirectory (), FILE_VARIANCE_LABELS);
                firstTime = false;
            }

            if ((labelsFile = fopen (labelsPathName, mode)) != NULL)
            {
                fprintf (labelsFile, "%s\n", varianceFileName);
                fclose (labelsFile);
            }
            else
                perror (labelsPathName);
        }
    }
    else
        perror (variancePathName);
}


/*============================================================================
 *
 * Name:: UpdateXtransposeX
 *
 * Description::
 *     Updates M=XT*X (X-transpose x X ) Here: X[genes][Jmax], M[Jmax][Jmax]
 *     to reflect the deletion of the genes (rows) specified.
 *
 * Inputs::
 *     X       - X[...][Jmax] - Original gene expression matrix.
 *     Jmax    - Length of each gene (number of columns).
 *     count   - Number of deleted genes
 *     geneMap - Vector[0..count-1] - Indices of deleted rows (genes).
 *
 * Inputs/Outputs::
 *     M       - M[Jmax][Jmax]
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

static void UpdateXtransposeX (
  matrix_t *M,
  matrix_t *X,
  int Jmax,
  int count,
  const ivector_t *geneMap)
{
    register int i;

    assert (M != NULL);
    assert (X != NULL);
    assert (geneMap != NULL);

    for (i = 0; i < count; i++)
    {
        register int j;
        int gene;

        gene = gsl_vector_int_get (geneMap, i);
        for (j = 0; j < Jmax; j++)
        {
            register int k;
            double gj = gsl_matrix_get (X, gene, j);

            for (k = 0; k < j; k++)
            {
                double v = gj * gsl_matrix_get (X, gene, k);

                gsl_matrix_set (M, j, k, gsl_matrix_get (M, j, k) - v);
                gsl_matrix_set (M, k, j, gsl_matrix_get (M, k, j) - v);
            }
            gsl_matrix_set (M, j, j, gsl_matrix_get (M, j, j) - gj * gj);
        }
    }
}

