/*=================== FILE ABSTRACT ==========================================
 *
 * Synopsis::         Source for shared private routines
 *
 * File Name::        %M%
 * Version::          %I%
 * Type::             source
 * Date Completed::   %G%
 *
 * Environment::
 *     Language::     ANSI C
 *     Computer::     Compaq/Alpha
 *     OS::           DigitalUNIX
 *     OSVersion::    4.0
 *
 * Module Information::
 *     Owner::        MDACC
 *     Department::   Biostatistics
 *     Researcher::   Kim-Anh Do
 *     Application::  geneclust
 *
 *============================================================================*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <assert.h>
#include <gsl/gsl_linalg.h>
#include "Utility.h"
#include "GeneClust.h"


/*
 * Protected Routines
 */

/*============================================================================
 *
 * Name:: DataDump
 *
 * Description::
 *     Stores the matrix with dimensions maxI by maxJ into a file for input
 *     to S-plus (column major order).
 *
 * Inputs::
 *     EZ       - EZ [...][maxJ] - Matrix to be stored
 *     maxJ     - Number of columns
 *     maxI     - Number of rows
 *     geneMap  - Vector [0..maxI-1] of gene (row) numbers.
 *                Null implies identity mapping.
 *     filename - Name of the file where the data should be stored
 *
 * Outputs::
 *     (None)
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

void DataDump (
  matrix_t *EZ,
  int maxJ,
  int maxI,
  const ivector_t *geneMap,
  const char *filename)
{
    FILE *file;
    register int j;
    register int i;

    assert (EZ != NULL);
    assert (maxI > 0);
    assert (maxJ > 0);
    assert (filename != NULL);

    if ((file = fopen (filename, "w")) == NULL)
    {
        ErrMessage_t errMessage;

        sprintf (errMessage, "DataDump(\"%s\"): %s\n",
                 filename,
                 strerror (errno));
        Abend (errMessage);
        /*NOTREACHED*/
    }

    for (j = 0; j < maxJ; j++)
    {
        for (i = 0; i < maxI; i++)
            fprintf (file, "%8.6f\n",
	        gsl_matrix_get (EZ, geneMap ? gsl_vector_int_get (geneMap, i) : i, j));
    }

    fclose (file);
}


/*============================================================================
 *
 * Name:: InitMatrix
 *
 * Description::
 *     Initializes a 2D matrix of double to zero
 *
 * Inputs::
 *     maxI - Number of rows
 *     maxJ - Number of columns
 *
 * Outputs::
 *     M    - M[maxI][maxJ] - Matrix to be initialized
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

void InitMatrix (matrix_t *M, int maxI, int maxJ)
{
    register int i;
    register int j;

    assert (M != NULL);
    assert (maxI > 0);
    assert (maxJ > 0);

    for (i = 0; i < maxI; i++)
    {
        for (j = 0; j < maxJ; j++)
            gsl_matrix_set (M, i, j, 0.0);
    }
}


/*============================================================================
 *
 * Name:: InitVector
 *
 * Description::
 *     Initializes a vector of double to zero
 *
 * Inputs::
 *     maxI - Number of elements
 *
 * Outputs::
 *     V    - V[maxI] - Vector to be initialized
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

void InitVector (vector_t *V, int maxI)
{
    register int i;

    assert (V != NULL);
    assert (maxI > 0);

    for (i = 0; i < maxI; i++)
        gsl_vector_set (V, i, 0.0);
}


/*============================================================================
 *
 * Name:: CenterData
 *
 * Description::
 *     Centers each row so that its mean is 0.0
 *
 * Inputs::
 *     N  - Number of rows
 *     p  - Number of columns
 *
 * Outputs::
 *     M  - Matrix whose elements are to be ceneter.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

void CenterData (matrix_t *M, int N, int p)
{
    register int i;
    register int j;
    double mean;
    double sum;

    assert (M != NULL);
    assert (N > 0);
    assert (p > 0);

    for (i = 0; i < N; i++)
    {
        sum = 0.0;
        for (j = 0; j < p; j++)
            sum += gsl_matrix_get (M, i, j);
        mean = sum / p;
        for (j = 0; j < p; j++)
            gsl_matrix_set (M, i, j, gsl_matrix_get (M, i, j) - mean);
    }
}


/*============================================================================
 *
 * Name:: StoreMatrixAsRowMajor
 *
 * Description::
 *     Store an existing matrix in a row-major order vector.
 *
 * Inputs::
 *     M - Matrix to be stored.
 *     N - Number of rows.
 *     p - Number of columns.
 *
 * Outputs::
 *     V - Vector in row-major order.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

void StoreMatrixAsRowMajor (matrix_t *M, double *V, int N, int p)
{
    register int i;
    register int j;

    assert (M != NULL);
    assert (V != NULL);

    for (i = 0; i < N; i++)
    {
        for (j = 0; j < p; j++)
        {
            V[i * p + j] = gsl_matrix_get (M, i, j);
        }
    }
}


/*============================================================================
 *
 * Name:: StoreMatrixAsColMajor
 *
 * Description::
 *     Store an existing matrix in a column-major order vector.
 *
 * Inputs::
 *     M - Matrix to be stored.
 *     N - Number of rows.
 *     p - Number of columns.
 *
 * Outputs::
 *     V - Vector in column-major order.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

void StoreMatrixAsColMajor (matrix_t *M, double *V, int N, int p)
{
    register int i;
    register int j;

    assert (M != NULL);
    assert (V != NULL);

    for (j = 0; j < p; j++)
    {
        for (i = 0; i < N; i++)
        {
            V[i + j * N] = gsl_matrix_get (M, i, j);
        }
    }
}


/*============================================================================
 *
 * Name:: FetchMatrixAsRowMajor
 *
 * Description::
 *     Restore a matrix in row-major order from a vector.
 *
 * Inputs::
 *     V - Vector in row-major order.
 *     N - Number of rows.
 *     p - Number of columns.
 *
 * Outputs::
 *     M - Matrix to be retrieved.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

void FetchMatrixAsRowMajor (double *V, matrix_t *M, int N, int p)
{
    register int i;
    register int j;

    assert (M != NULL);
    assert (V != NULL);

    for (i = 0; i < N; i++)
    {
        for (j = 0; j < p; j++)
        {
            gsl_matrix_set (M, i, j, V[i * p + j]);
        }
    }
}


/*============================================================================
 *
 * Name:: FetchMatrixAsColMajor
 *
 * Description::
 *     Restore a matrix in column-major order from a vector.
 *
 * Inputs::
 *     V - Vector in column-major order.
 *     N - Number of rows.
 *     p - Number of columns.
 *
 * Outputs::
 *     M - Matrix to be retrieved.
 *
 * Return Value::
 *     (None)
 *
 *============================================================================*/

void FetchMatrixAsColMajor (double *V, matrix_t *M, int N, int p)
{
    register int i;
    register int j;

    assert (M != NULL);
    assert (V != NULL);

    for (j = 0; j < p; j++)
    {
        for (i = 0; i < N; i++)
        {
            gsl_matrix_set (M, i, j, V[i + j * N]);
        }
    }
}

gsl_matrix *
dmatrix (int rl, int rh, int cl, int ch)
{
	if (rl != 0 || cl != 0) {
		fprintf (stderr, "dmatrix calling with non-zero rl or cl\n");
		abort ();
	}
	return gsl_matrix_alloc (rh+1, ch+1);
}

void
free_dmatrix (gsl_matrix *m, int rl, int rh, int cl, int ch)
{
	gsl_matrix_free (m);
}

gsl_vector_int *
ivector (int rl, int rh)
{
	if (rl != 0) {
		fprintf (stderr, "ivector calling with non-zero rl\n");
		abort ();
	}
	return gsl_vector_int_alloc (rh+1);
}

void
free_ivector (gsl_vector_int *m, int rl, int rh)
{
	gsl_vector_int_free (m);
}


gsl_vector *
dvector (int rl, int rh)
{
	if (rl != 0) {
		fprintf (stderr, "dvector calling with non-zero rl\n");
		abort ();
	}
	return gsl_vector_alloc (rh+1);
}

void
free_dvector (gsl_vector *v, int rl, int rh)
{
	gsl_vector_free (v);
}


void
handler (const char *reason, const char *file, int line, int gsl_errno)
{
    fprintf (stderr, "GSL Error: %s at line %d in %s\n", reason, line, file);
}

void
choldc (gsl_matrix *a, int n, gsl_vector *p)
{
    fprintf (stderr, "> choldc\n");
    gsl_linalg_cholesky_svx (a, p);
    fprintf (stderr, "< choldc\n");
}

void
svdcmp (gsl_matrix *mat, gsl_matrix *a, int n, int m, gsl_vector *d, gsl_matrix *v)
{
    gsl_error_handler_t *old_handler;

#ifdef TRACEFNS
    fprintf (stderr, "> svdcmp\n");
#endif
    old_handler = gsl_set_error_handler (&handler);
    gsl_matrix_memcpy (a, mat);
    if (gsl_linalg_SV_decomp_jacobi (a, v, d)) {
        gsl_vector *w = gsl_vector_alloc (n);
        gsl_matrix_memcpy (a, mat);
	if (gsl_linalg_SV_decomp (a, v, d, w))
	    fprintf (stderr, "ERROR: gsl_linalg_SV_decomp failed\n");
	gsl_vector_free (w);
    }
    gsl_set_error_handler (old_handler);
#ifdef TRACEFNS
    fprintf (stderr, "< svdcmp\n");
#endif
}

char **cvector (int n)
{
    char **s;
    register int i;

    s= (char **)calloc (n, sizeof(char *));
    if (s == NULL)
        Abend ("allocation failure 1 in cvector");

    for (i = 0; i < n; i++)
    {
        s[i] = (char *)calloc (CVECTOR_ELEM_LEN+1, sizeof(char));
        if (s[i] == NULL)
            Abend ("allocation failure 2 in cvector");
    }

    return s;
}


void free_cvector (char **s, int n)
{
    register int i;

    for (i = n-1; i >= 0; i--)
    {
        if (s[i] != NULL)
        {
            free ((char *) s[i]);
            s[i] = NULL;
        }
    }

    free ((char *)s);
    s = NULL;
}


void **pvector (int n)
{
    char **v;

    v= (char **)calloc ((size_t) n, sizeof(void *));
    if (v == NULL)
        Abend ("allocation failure in pvector()");

    return (void **)v;
}


/*ARGSUSED*/
void free_pvector (void **v, int n)
{
    free ((char *)v);
    v = NULL;
}

